//CPU handler  starts/stops cpu etc
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <process.h>

#include "dd.h"

#include "types.h"

#include "message.h"
#include "file.h"

#include "keyb.h"

#include "m6502.h"

#include "nes.h"
#include "nesvideo.h"
#include "slist.h"


byte CPUpaused=0; //flag for when cpu is paused
byte CPURunning=0; //flag for when cpu is running
int CPUtickframe=0; //tick a frame at a time

volatile int cycles;

void startframe();
void startvblank();

#include "timing.h"

extern volatile byte frame;

//are we currently in a virtual frame ? (emulating)
volatile byte inemu=0;

void m_showmessages();

void trapfunc()
{
 msg.printf(1,"addr: %X",m6502pc);
}

void cpuabort()
{
 word addr=m6502pc;
 msg.error("Unrecognized opcode %X at addr: %X",
    (addr<0x8000) ? ram[addr] : m6502rom[addr],addr);
// msg.error("ram=%X rom=%X",m6502ram,m6502rom);
//  msg.error("cycles: %d %d",m6502clockticks,VBLANKCYCLES);
 m_stop();
 m_showmessages();
}


//get current virtual scanline being draw
int getscanline()
{
// return frame ? ((m6502clockticks+HBLANKCYCLES/2)/HBLANKCYCLES) : 0;
 if (!frame) return 0;
 return m6502clockticks/HBLANKCYCLES;
}

byte IRQline=0xC0; //line to do IRQ

//ticked (once per virtual frame)
void tickemu()
{
 if (inemu) return;
 inemu=1;

 //reset event list
 sl.reset();

 //execute during vblank
 startvblank();
 m6502clockticks=0;
 if (ram[0x2000]&0x80) m6502nmi(); //generate vblank?
 if (m6502exec(VBLANKCYCLES)!=0x10000) {cpuabort(); inemu=0; return;}

 //execute during virtual frame...
 startframe();
 m6502clockticks=HBLANKCYCLES*8; //reset clock ticks

 //if (IRQline>FRAMELINES || IRQline<8)
  { //dont bother with IRQ
   if (m6502exec(FRAMECYCLES-m6502clockticks)!=0x10000) {cpuabort(); inemu=0; return;}
  }/* else
  { //do IRQ
   int cyclesbeforeIRQ=IRQline*HBLANKCYCLES;
   if (m6502exec(cyclesbeforeIRQ)!=0x10000) {cpuabort(); inemu=0; return;}
   m6502int();
   if (m6502exec(FRAMECYCLES-cyclesbeforeIRQ)!=0x10000) {cpuabort(); inemu=0; return;}
  }  */


 //create context list based on event list
 sc.create(sl);

 //done
 cycles+=CYCLESPERTICK;
 inemu=0;
}


void m_execute()
{
 if (!romloaded) {msg.error("ROM not loaded"); return;}
 if (CPURunning) {msg.error("CPU already running"); return;}
 if (resetNEShardware()!=0)
   { msg.error("Unable to reset NES hardware"); return;}

 m6502trap=trapfunc;
 IRQline=0xFF; //no irq
 msg.printf(3,"CPU emulation started");
 CPURunning=1;
 CPUpaused=0;
}

void m_stop()
{
 if (!romloaded) return;
 CPUpaused=0;
 if (CPURunning) msg.printf(3,"CPU emulation stopped");
 CPURunning=0;
 inemu=0;

 //copy battery backed ram
 if (batterymem) batterymem->copyfromRAM();
}

void m_resume()
{
 if (!romloaded) {msg.error("ROM not loaded"); return;}
 if (CPURunning) {msg.error("CPU already running"); return;}
 if (!CPUpaused) {msg.error("CPU not paused"); return;}

 msg.printf(3,"CPU emulation resumed");
 CPURunning=1;
 CPUpaused=0;
}

void m_pause()
{
 if (!romloaded) return;
 if (!CPURunning) {msg.error("CPU not running"); return;}
 CPUpaused=1;
 CPURunning=0;
 msg.printf(3,"CPU emulation paused");
}

void m_advanceframe()
{
 if (!romloaded) return;
 if (!CPURunning && !CPUpaused) m_execute();
 if (!CPUpaused) m_pause();
 if (!CPUpaused) return;
 CPUtickframe++;
}

//-----------------------------------------------------------
//-----------------------------------------------------------



