locals

.486p
.model flat,c



;************************************************************************
;*                                                                      *
;* Equates, publics and options                                         *
;*                                                                      *
;************************************************************************

TASM equ 1

; reset the memory PC before each read/write
;RESETPC equ 1
;LASTJSR equ 1
;MSVC equ 1



ifdef TASM
		public  m6502clockticks
		public  m6502ram,m6502rom,m6502lastjsr
		public  m6502MemoryRead
		public  m6502MemoryWrite
		public  m6502pc
		public  m6502af
		public  m6502x
		public  m6502y
		public  m6502s
		public	cyclesRemaining
		public	syncCycle
      public inNmi

		public  m6502nmi
		public  m6502int
		public  m6502GetInfo
		public  m6502exec
		public  m6502reset
endif

MAJOR_REV       equ     1       ; Major revision
MINOR_REV       equ     0       ; Minor revision
MULTI_CPU       equ     0       ; Enable multi-CPU support
INST_COUNT	equ	0	; We're counting down by instructions
CYCLE_COUNT	equ	1	; We're counting down by virtual clock time

ifdef MULTI_CPU


ifdef TASM
		public  m6502GetContextSize
		public  m6502SetContext
		public  m6502GetContext
endif

endif

; M6502.ASM OPTIONS:

; Comment this out to not include timing information in emulation

TICKS           equ     1       ; Uncomment if we want a clock count

; Comment this out if you don't need to monitor the SYNC line

;SYNC_WANTED    equ     1       ; Uncomment if we want SYNC output

; Comment this out when you have released your code. This does a check to
; make sure that each instruction is fetched within the 64K segment in
; question. It will slow execution significantly, so comment it back out when
; you're ready to release your emulation.

;BOUNDS_CHECK    equ     1      ; Uncomment if you want bounds checking for
				; address space. Comment this back out when
				; your emulation is working properly

; This controls what the value passed to m6502exec actually means. If this
; is set to the INST_COUNT method, then it will consider the value input to be
; the number of virtual instructions left to execute. If this is set to
; CYCLE_COUNT then it is considered to be the number of virtual cycles left
; to execute.
;
; Be warned that changing this from INST_COUNT to CYCLE_COUNT will give
; approximately a 9% performance hit. If the image being emulated doesn't
; need exact timing, then keep it at INST_COUNT and approximate the instructions
; that pass between each cycle.

CYCLE_METHOD	equ	CYCLE_COUNT ; Count instructions

; ** DO NOT MUCK WITH THESE LABELS **

; FLAG REGISTER BIT FIELD NAMES
;

CARRY           equ     01h     ; Carry flag
ZERO            equ     02h     ; Zero flag
INTERRUPT       equ     04h     ; Interrupt flag
DECIMAL         equ     08h     ; That dreaded decimal mode
BREAK           equ     10h     ; Break flag

RESERVED        equ    20h      ;always 1

; Note that 20h is not used

OVERFLOW        equ     40h     ; Overflow flag
SIGN            equ     80h     ; Sign flag

;************************************************************************
;*                                                                      *
;* Macros                                                               *
;*                                                                      *
;************************************************************************

; Used when fetching instructions. It fetches the byte into AL from what is
; pointed to by ESI and leaves all other registers undisturbed

FetchInstructionByte    macro

; If you're using a 486, use the LODSW. If you're using a Pentium, use the
; instructions following it. It gives about a 7% increase in speed.

;		lodsb                   ; Fetch the next instruction into AL

		mov	al, [esi]
		inc	esi

		endm

; Used when fetching part of instructions (addresses, offsets, etc...). It
; fetches the two byte data into AX from ESI and leaves the other registers
; undisturbed

FetchInstructionWord    macro

; If you're using a 486, use the LODSW. If you're using a Pentium, use the
; instructions following it. It gives about a 7% increase in speed.

;		lodsw                   ; Fetch the next instruction word into AX

		mov	ax, [esi]
		add	esi, 2

		endm

; Used when attempting to do a memory read from any address. The read address
; is expected to be in AX - the 6502's 64K memory space. Any fixups are done
; by routines that may have it covered. AL=Byte that just got read.

ReadMemoryByte  macro
      local isROM, isRAM, done

      ;cmp  ah,80h ;see if high portion of addr space?
      ;jae   isROM
      test ah,ah
      js isROM

      cmp  ah,20h ;see if low 8k of RAM
      jb   isRAM
      call   memoryread ;call memory read function
      jmp   done

isRAM: and ah,7h     ;force mirroring to lower 2K
       mov al,[ebp+eax]
       jmp done
isROM: mov al,[edi+eax] ;read the rom image directly
done:  and     eax, 0ffh       ; Only the lower byte

		endm


; Used when writing a memory byte to an address space. The address is
; expected to be in AX, and the byte to write in DL. Any fixups are done by
; the routines that have covered it.

WriteMemoryByte macro
      local notRAM, noReloadPc, done

      cmp  ah,20h ;see if low 8k of RAM
      jae  notRAM
      and  ah,7h     ;force mirroring to lower 2K
      mov [ebp+eax],dl ;write it directly
      jmp  done
notRAM: call memorywrite ; do memory write function
done:
		endm


 ;;converts 16-bit PC -> 32-bit pointer
naturalizeESI macro
   local isrom, done,notlowerram
     test si,si ;;;8000h
     js  isrom
;     jmp invalidInsByte ; dont let them run code <8000
     cmp si,2000h
     jae notlowerram
     and si,7FFh
notlowerram:
     add esi,ebp
     mov [m6502esibase],ebp
;   jmp invalidInsByte
     jmp done
isrom:
     add esi,edi
     mov [m6502esibase],edi
done:
;;  add esi,edi
     endm



 ;converts 32-bit pointer to 16-bit PC
denaturalizeESI macro
     sub esi,[m6502esibase]
  ;;;;;  sub esi,edi
          endm



; This is the macro to use to add something to our clock.

Clocktime       macro   addedTime
		endm

; This macro is used at the end of an emulated function. It will return to the
; main execution loop to get the next instruction. IT IS SUPPOSED TO BE A
; JUMP AND NOT A CALL!

ExecuteNextInstruction  macro	clkTiming

      mov eax,clkTiming
      add m6502clockticks,eax ; add to total counter
		sub cyclesRemaining,eax ; Subtract the virtual clock ticks
		jle instReturn

		FetchInstructionByte
		jmp     dword ptr m6502regular[eax*4] ; Jump to our instruction!

		endm

; This macro is used at the beginning of every emulated function, and is
; different between assemblers

Alignment       macro

    align
		endm

; Macro used when all flags need to be set (value in AL)

FlagSet         macro
		and     dh, 07dh        ; Knock out zero and SIGN
		xor     ah, ah          ; Zero this just in case
		or      dh, byte ptr flagTable[eax] ; Go get our flag

		endm

FlagSetX        macro

		and     dh, 07dh        ; Knock out zero and SIGN
		or      dh, byte ptr flagTable[ebx] ; Go get our flag

		endm

FlagSetY        macro

		and     dh, 07dh        ; Knock out zero and SIGN
		or      dh, byte ptr flagTable[ecx] ; Go get our flag

		endm

; Macro used when all flags need to be set from value in ACCUMULATOR

FlagSetAcc      macro

		xor	eax, eax		; Zero this for offset
		mov     al, dl          ; Move accumulator into AL
		and     dh, 07dh        ; Knock out ZERO and SIGN
		or      dh, byte ptr flagTable[eax] ; OR it in!

		endm

; Macro used when flags need to be set EXCEPT overflow (value in AL)

FlagSetNoOv     macro

		and	dh, 07dh	; No sign - just zero
		xor	ah, ah		; Zero this for lookup
		or	dh, byte ptr flagNoOvTable[eax]

		endm

; Macro used when flags need to be set EXCEPT overflow value in ACCUMULATOR

FlagSetNoOvAcc  macro
		local   l1

		and     dh, 07dh        ; Now we're zero
		xor	ah, ah		; Zero this for table lookup
		mov	al, dl		; Make it the accumulator lookup
		or	dh, byte ptr flagNoOvTable[eax]

		endm




; This is the ADC macro used by various instances of ADC. The value to add
; to the accumulator is in AL.

AdcMacro        macro
		local   doDecAdc, noCarrySetDecAdc, skipTensUpgrade
		local   noCarryDecAdc, setFlags, adcOut
      local noover
;;;		local   addIt, noCarry, noSign, noAdcCarry, adcCont

		test    dh, DECIMAL     ; Are we in decimal mode?
		jnz     doDecAdc

    and dh,0FFh-OVERFLOW ;no overflow
    rcr dh,1 ;get carry
    adc dl,al ;add with carry
    jno noover
    rcl dh,1 ;get carry from addition
    or  dh,OVERFLOW
    FlagSetAcc
    jmp adcOut
noover:
    rcl dh,1 ;get carry from addition
    FlagSetAcc
    jmp adcOut

; Here we need to add two decimal #'s together

doDecAdc:
		push    ecx
		push    ebx

		xor     ebx, ebx          ; Zero BX for later

; Add the first nibble together

		mov     ch, al          ; Source nibble
		mov     cl, dl          ; Destination nibble
		and     cx, 0f0fh       ; Only lower nibble allowed!
		add     cl, ch          ; Add them together
		test    dh, CARRY       ; Carry set?
		jz      noCarrySetDecAdc
		inc     cl              ; Increment!
noCarrySetDecAdc:
		cmp     cl, 0ah         ; Are we up there?
		jb      skipTensUpgrade
		mov     bh, 10h         ; Indicate we need to INC the upper nibble
		sub     cl, 0ah         ; Just the 10's location

; At this point, CL contains the lower BCD, and BH is non-zero if we've
; overflowed

skipTensUpgrade:
		mov     bl, cl          ; Put it in BL for later adding
		mov     ch, al          ; Source nibble
		mov     cl, dl          ; Source nibble
		and     cx, 0f0f0h      ; Only upper nibble this time
		add     cl, ch          ; Get it!
		add     cl, bh          ; Add in our lower BCD (if applicable)
		and     dh, 0beh        ; Clear carry and overflow
		cmp     cl, 0a0h        ; Are we too big?
		jb      noCarryDecAdc   ; Nope!

		or      dh, CARRY       ; Set virtual carry
		sub     cl, 0a0h        ; Make it be in range
noCarryDecAdc:
		or      bl, cl          ; Put our new value in BL
		mov     al, bl          ; Put it in AL

setFlags:
		pop     ebx              ; Restore BX

		mov     dl, al          ; Put it in accumulator
		FlagSet
		pop     ecx              ; Restore CX
adcOut:

		endm


; This is the SBC macro used by various instances of SBC. The value to subtract
; to the accumulator is in AL.

SbcMacro        macro
		local   noCarry, noSign, noSbcCarry, sbcCont
		local   sbcHandleDec, noSbcDec, noUnderFlow, noUnderFlowMSB
		local   skipIt, sbcOut

		push    ecx              ; Save this away for later

		test    dh, DECIMAL     ; Are we in decimal mode?
		jnz     sbcHandleDec

		xor     al, 0ffh        ; Invert what we're adding
		mov     ch, al          ; Store our original data
		add     al, dl          ; Add in A
		adc     ah, 0           ; If we have a carry...
		test    dh, CARRY       ; Carry set?
		jz      noCarry
		inc     ax              ; Increment A
noCarry:

; Let's check out and see if we've overflowed

		and     dh, 0bfh        ; No overflow
		xor     ch, dl          ; XOR A with our data
		test    ch, 80h         ; Sign extended?
		jnz     noSign

		mov     ch, dl          ; Get A
		xor     ch, al          ; XOR It with our new resultant
		jz      noSign

		or      dh, OVERFLOW    ; Indicate we've overflowed

noSign:
		mov     dl, al          ; Put A away for later

		or      ah, ah          ; Shall we carry?
		jz      noSbcCarry

		or      dh, 1           ; We've got a carry!
		jmp     short sbcOut

noSbcCarry:
		and     dh, 0ffh - CARRY ; No carry! We didn't overflow
		jmp     short sbcOut

; Here we handle decimal mode

sbcHandleDec:
		push    ebx              ; Save it because we wipe it out
		xor     ebx, ebx          ; Zero our borrow for later

		mov     ch, dl          ; A register in CH
		mov     cl, al          ; Data in CL
		and     cx, 0f0fh       ; Lower nibble only this time

		sub     ch, cl          ; Get our first digit's results
		test    dh, 01h         ; Carry set? (borrow)?
		jnz     noSbcDec
		dec     ch              ; Back up one for carry!
noSbcDec:
		cmp     ch, 0ah         ; Have we overflowed? (underflowed?)
		jb      noUnderFlow
		add     ch, 0ah         ; Add it in
		mov     bh, 10h         ; We take it from the most significant BCD

noUnderFlow:
		mov     bl, ch          ; Put final in BL

; We now have the lower BCD in BL, and any borrow in BH

		and     dh, 0beh        ; Clear carry and overflow flag

		mov     ch, dl          ; A register in CH
		mov     cl, al          ; Data in CL
		and     cx, 0f0f0h      ; Upper nibble this time
		sub     ch, cl          ; Get our second digit's results
		sub     ch, bh          ; Subtract our borrow (if applicable)
		cmp     ch, 0a0h        ; Did we underflow?
		jae     noUnderFlowMSB

		or      dh, 01h         ; Set carry if we underflowed
		jmp     short skipIt

noUnderFlowMSB:
		add     ch, 0a0h        ; Put it back to normal

skipIt:
		or      bl, ch          ; Store it in BL

; Need to put it back into A, here!

		mov     al, bl
		mov     dl, bl
		pop     ebx
sbcOut:
		FlagSet
		pop     ecx
		endm


;************************************************************************
;*                                                                      *
;* Emulation data                                                       *
;*                                                                      *
;************************************************************************


_DATA           segment         dword public 'DATA'

Alignment

; CONTEXT CONTEXT CONTEXT CONTEXT CONTEXT CONTEXT CONTEXT CONTEXT CONTEXT

ContextBegin:

m6502ram  dd 0
m6502rom  dd 0
m6502esibase dd 0

m6502MemoryRead dd     0       ; Offset of memory read structure array
m6502MemoryWrite dd    0       ; Offset of memory write structure array
m6502clockticks dd     0       ; # Of clock tips that have elapsed
m6502af        dw      0       ; A register and flags
m6502pc        dw      0       ; PC Register
m6502x         db      0       ; X register
m6502y         db      0       ; Y register
m6502s         db      0       ; S register
syncCycle      db      0       ; Non-zero if this is a sync cycle
inNmi		db	0	; Non-zero if we're in an NMI

m6502lastjsr dw 0 ; last address of jsr taken

Alignment

ContextEnd:


; CONTEXT CONTEXT CONTEXT CONTEXT CONTEXT CONTEXT CONTEXT CONTEXT CONTEXT

; This is general purpose non-persistent data

Alignment

cyclesRemaining dd      0       ; # Of virtual CPU cycles left to execute

Alignment

; This is the jump table for the regular instructions

m6502regular  LABEL DWORD
      dd      inst00                    ; 00h - BRK
		dd      inst01                    ; 01h - ORA A, INDX
		dd      invalidInsByte            ; 02h - Invalid
		dd      invalidInsByte            ; 03h - Invalid
		dd      invalidInsByte	          ; 04h - Invalid
		dd      inst05                    ; 05h - ORA $xx
		dd      inst06                    ; 06h - ASL $xx
		dd      invalidInsByte            ; 07h - Invalid
		dd      inst08                    ; 08h - PHP
		dd      inst09                    ; 09h - ORA #xxh
		dd      inst0a                    ; 0ah - ASL A
		dd      invalidInsByte            ; 0bh - Invalid
		dd	invalidInsByte		  ; 0ch - Invalid
		dd      inst0d                    ; 0dh - ORA $xxxxh
		dd      inst0e                    ; 0eh - ASL $xxxx
		dd      invalidInsByte            ; 0fh - Invalid
		dd      inst10                    ; 10h - BPL $xxxx
		dd      inst11		          ; 11h - ORA ($xx), Y
		dd      invalidInsByte            ; 12h - Invalid
		dd      invalidInsByte            ; 13h - Invalid
		dd      invalidInsByte            ; 14h - Invalid
		dd      inst15                    ; 15h - ORA $xx, X
		dd      inst16		          ; 16h - ASL $xx, X
		dd      invalidInsByte            ; 17h - Invalid
		dd      inst18                    ; 18h - CLC
		dd      inst19		          ; 19h - ORA $xxxx, Y
		dd      invalidInsByte            ; 1ah - Invalid
		dd      invalidInsByte            ; 1bh - Invalid
		dd      invalidInsByte            ; 1ch - Invalid
		dd      inst1d                    ; 1dh - ORA $xxxx, X
		dd      inst1e		          ; 1eh - ASL $xxxx, X
		dd      invalidInsByte            ; 1fh - Invalid
		dd      inst20                    ; 20h - JSR $xxxx
		dd      inst21		          ; 21h - AND ($xx), X
		dd      invalidInsByte            ; 22h - Invalid
		dd      invalidInsByte            ; 23h - Invalid
		dd      inst24                    ; 24h - BIT $xx
		dd      inst25                    ; 25h - AND $xx
		dd      inst26                    ; 26h - ROL $xx
		dd      invalidInsByte            ; 27h - Invalid
		dd      inst28                    ; 28h - PLP
		dd      inst29                    ; 29h - AND #xxh
		dd      inst2a                    ; 2ah - ROL A
		dd      invalidInsByte            ; 2bh - Invalid
		dd      inst2c                    ; 2ch - BIT $xxxx
		dd      inst2d                    ; 2dh - AND $xxxx
		dd      inst2e		          ; 2eh - ROL $xxxx
		dd      invalidInsByte            ; 2fh - Invalid
		dd      inst30                    ; 30h - BMI $xxxx
		dd      inst31		          ; 31h - AND ($xx), Y
		dd      invalidInsByte            ; 32h - Invalid
		dd      invalidInsByte            ; 33h - Invalid
		dd      invalidInsByte            ; 34h - Invalid
		dd      inst35		          ; 35h - AND $xx, X
		dd      inst36		          ; 36h - ROL $xx, X
		dd      invalidInsByte            ; 37h - Invalid
		dd      inst38                    ; 38h - SEC
		dd      inst39		          ; 39h - AND $xxxx, Y
		dd      invalidInsByte            ; 3ah - Invalid
		dd      invalidInsByte            ; 3bh - Invalid
		dd      invalidInsByte            ; 3ch - Invalid
		dd      inst3d                    ; 3dh - AND $xxxx, X
		dd      inst3e		          ; 3eh - ROL $xxxx, X
		dd      invalidInsByte            ; 3fh - Invalid
		dd      inst40                    ; 40h - RTI
		dd      inst41		          ; 41h - EOR ($xx), X
		dd      invalidInsByte            ; 42h - Invalid
		dd      invalidInsByte            ; 43h - Invalid
		dd      invalidInsByte            ; 44h - Invalid
		dd      inst45                    ; 45h - EOR $xx
		dd      inst46                    ; 46h - LSR $xx
		dd      invalidInsByte            ; 47h - Invalid
		dd      inst48                    ; 48h - PHA
		dd      inst49                    ; 49h - EOR #$xx
		dd      inst4a                    ; 4ah - LSR A
		dd      invalidInsByte            ; 4bh - Invalid
		dd      inst4c                    ; 4ch - JMP $xxxx
		dd      inst4d		          ; 4dh - EOR $xxxx
		dd      inst4e                    ; 4eh - LSR $xxxx
		dd      invalidInsByte            ; 4fh - Invalid
		dd      inst50                    ; 50h - BVC $xxxx
		dd      inst51		          ; 51h - EOR ($xx), Y
		dd      invalidInsByte            ; 52h - Invalid
		dd      invalidInsByte            ; 53h - Invalid
		dd      invalidInsByte            ; 54h - Invalid
		dd      inst55		          ; 55h - EOR $xx, X
		dd      inst56		          ; 56h - LSR $xx, X
		dd      invalidInsByte            ; 57h - Invalid
		dd      inst58		          ; 58h - CLI
		dd      inst59		          ; 59h - AND $xxxx, Y
		dd      invalidInsByte            ; 5ah - Invalid
		dd      invalidInsByte            ; 5bh - Invalid
		dd      invalidInsByte            ; 5ch - Invalid
		dd      inst5d                    ; 5dh - EOR $xxxx, X
		dd      inst5e		          ; 5eh - LSR $xxxx, X
		dd      invalidInsByte            ; 5fh - Invalid
		dd      inst60                    ; 60h - RTS
		dd      inst61		          ; 61h - ADC ($xx,X)
		dd      invalidInsByte            ; 62h - Invalid
		dd      invalidInsByte            ; 63h - Invalid
		dd      invalidInsByte            ; 64h - Invalid
		dd      inst65                    ; 65h - ADC ($02)
		dd      inst66                    ; 66h - ROR $xxh
		dd      invalidInsByte            ; 67h - Invalid
		dd      inst68                    ; 68h - PLA
		dd      inst69                    ; 69h - ADC #xxh
		dd      inst6a                    ; 6ah - ROR A
		dd      invalidInsByte            ; 6bh - Invalid
		dd      inst6c		          ; 6ch - JMP ($xxxx)
		dd      inst6d                    ; 6dh - ADC $xxxx
		dd      inst6e                    ; 6eh - ROR $xxxxh
		dd      invalidInsByte            ; 6fh - Invalid
		dd      inst70                    ; 70h - BVS $xxxx
		dd      inst71                    ; 71h - ADC ($xx),Y
		dd      invalidInsByte            ; 72h - Invalid
		dd      invalidInsByte            ; 73h - Invalid
		dd      invalidInsByte            ; 74h - Invalid
		dd      inst75                    ; 75h - ADC $xx, X
		dd      inst76		          ; 76h - ROR $xx, X
		dd      invalidInsByte            ; 77h - Invalid
		dd      inst78		          ; 78h - SEI
		dd      inst79                    ; 79h - ADC $xxxx,Y
		dd      invalidInsByte            ; 7ah - Invalid
		dd      invalidInsByte            ; 7bh - Invalid
		dd      invalidInsByte            ; 7ch - Invalid
		dd      inst7d                    ; 7dh - ADC $xxxx,X
		dd      inst7e		          ; 7eh - ROR $xxxx, X
		dd      invalidInsByte            ; 7fh - Invalid
		dd      invalidInsByte            ; 80h - Invalid
		dd      inst81		          ; 81h - STA ($xx),Y
		dd      invalidInsByte            ; 82h - Invalid
		dd      invalidInsByte            ; 83h - Invalid
		dd      inst84                    ; 84h - STY $xx
		dd      inst85                    ; 85h - STA $xx
		dd      inst86                    ; 86h - STX $xx
		dd      invalidInsByte            ; 87h - Invalid
		dd      inst88                    ; 88h - DEY
		dd      invalidInsByte            ; 89h - Invalid
		dd      inst8a                    ; 8ah - TXA
		dd      invalidInsByte            ; 8bh - Invalid
		dd      inst8c                    ; 8ch - STY $xxxx
		dd      inst8d                    ; 8dh - STA $xxxx
		dd      inst8e                    ; 8eh - STX $xxxx
		dd      invalidInsByte            ; 8fh - Invalid
		dd      inst90                    ; 90h - BCC $xxxx
		dd      inst91                    ; 91h - STA $xx, Y
		dd      invalidInsByte            ; 92h - Invalid
		dd      invalidInsByte            ; 93h - Invalid
		dd      inst94		          ; 94h - STY $xx, X
		dd      inst95                    ; 95h - STA $00, X
		dd      inst96 		          ; 96h - STX $xx, Y
		dd      invalidInsByte            ; 97h - Invalid
		dd      inst98                    ; 98h - TYA
		dd      inst99                    ; 99h - STA $xxxx, Y
		dd      inst9a                    ; 9ah - TXS
		dd      invalidInsByte            ; 9bh - Invalid
		dd      invalidInsByte            ; 9ch - Invalid
		dd      inst9d                    ; 9dh - STA $xxxx, X
		dd      invalidInsByte            ; 9eh - Invalid
		dd      invalidInsByte            ; 9fh - Invalid
		dd      insta0                    ; a0h - LDY, ##h
		dd      insta1                    ; a1h - LDA ($xx), X
		dd      insta2                    ; a2h - LDX, ##h
		dd      invalidInsByte            ; a3h - Invalid
		dd      insta4                    ; a4h - LDY $xx
		dd      insta5                    ; a5h - LDA $xx
		dd      insta6                    ; a6h - LDX $xx
		dd      invalidInsByte            ; a7h - Invalid
		dd      insta8                    ; a8h - TAY
		dd      insta9                    ; a9h - LDA #xxh
		dd      instaa                    ; aah - TAX
		dd      invalidInsByte            ; abh - Invalid
		dd      instac                    ; ach - LDY $xxxx
		dd      instad                    ; adh - LDA $xxxx
		dd      instae                    ; aeh - LDX $xxxx
		dd      invalidInsByte            ; afh - Invalid
		dd      instb0                    ; b0h - BCS $xxxx
		dd      instb1                    ; b1h - LDA ($xx), Y
		dd      invalidInsByte            ; b2h - Invalid
		dd      invalidInsByte            ; b3h - Invalid
		dd      instb4                    ; b4h - LDY $xx,X
		dd      instb5                    ; b5h - LDA $xx,X
		dd      instb6		          ; b6h - LDX $xx,Y
		dd      invalidInsByte            ; b7h - Invalid
		dd      instb8		          ; b8h - CLV
		dd      instb9                    ; b9h - LDA $xxxxh, Y
		dd      instba		          ; bah - TSX
		dd      invalidInsByte            ; bbh - Invalid
		dd      instbc		          ; bch - LDY $xxxxh, X
		dd      instbd                    ; bdh - LDA $xxxxh, X
		dd      instbe                    ; beh - LDX $xxxxh, Y
		dd      invalidInsByte            ; bfh - Invalid
		dd      instc0                    ; c0h - CPY #xxh
		dd      instc1		          ; c1h - CMP ($xx), X
		dd      invalidInsByte            ; c2h - Invalid
		dd      invalidInsByte            ; c3h - Invalid
		dd      instc4                    ; c4h - CPY $xx
		dd      instc5                    ; c5h - CMP $xx
		dd      instc6                    ; c6h - DEC $xx
		dd      invalidInsByte            ; c7h - Invalid
		dd      instc8                    ; c8h - INY
		dd      instc9                    ; c9h - CMP #xxh
		dd      instca                    ; cah - DEX
		dd      invalidInsByte            ; cbh - Invalid
		dd      instcc		          ; cch - CPY $xxxx
		dd      instcd                    ; cdh - CMP $xxxx
		dd      instce                    ; ceh - DEC $xxxx
		dd      invalidInsByte            ; cfh - Invalid
		dd      instd0                    ; d0h - BNE $xxxx
		dd      instd1		          ; d1h - CMP ($xx), Y
		dd      invalidInsByte            ; d2h - Invalid
		dd      invalidInsByte            ; d3h - Invalid
		dd      invalidInsByte            ; d4h - Invalid
		dd      instd5                    ; d5h - CMP $xx, X
		dd      instd6                    ; d6h - DEC $xx, X
		dd      invalidInsByte            ; d7h - Invalid
		dd      instd8                    ; d8h - CLD
		dd      instd9		          ; d9h - CMP $xxxx, Y
		dd      invalidInsByte            ; dah - Invalid
		dd      invalidInsByte            ; dbh - Invalid
		dd      invalidInsByte            ; dch - Invalid
		dd      instdd		          ; ddh - CMP $xxxx, X
		dd      instde                    ; deh - DEC $xxxx, X
		dd      invalidInsByte            ; dfh - Invalid
		dd      inste0                    ; e0h - CPX #$xx
		dd      inste1		          ; e1h - SBC ($xx), X
		dd      invalidInsByte            ; e2h - Invalid
		dd      invalidInsByte            ; e3h - Invalid
		dd      inste4                    ; e4h - CPX $xx
		dd      inste5                    ; e5h - SBC $xx
		dd      inste6                    ; e6h - INC $xx
		dd      invalidInsByte            ; e7h - Invalid
		dd      inste8                    ; e8h - INX
		dd      inste9                    ; e9h - SBC #$xx
		dd      instea		          ; eah - NOP
		dd      invalidInsByte            ; ebh - Invalid
		dd      instec		          ; ech - CPX $xxxx
		dd      insted                    ; edh - SBC $xxxx
		dd      instee                    ; eeh - INC $xxxx
		dd      invalidInsByte            ; efh - Invalid
		dd      instf0                    ; f0h - BEQ $xxxx
		dd      instf1		          ; f1h - SBC ($xx), Y
		dd      invalidInsByte            ; f2h - Invalid
		dd      invalidInsByte            ; f3h - Invalid
		dd      invalidInsByte            ; f4h - Invalid
		dd      instf5                    ; f5h - SBC $xx, X
		dd      instf6                    ; f6h - INC $xx, X
		dd      invalidInsByte            ; f7h - Invalid
		dd      instf8                    ; f8h - SED
		dd      instf9		          ; f9h - SBC $xxxx, Y
		dd      invalidInsByte            ; fah - Invalid
		dd      invalidInsByte            ; fbh - Invalid
		dd      invalidInsByte            ; fch - Invalid
		dd      instfd                    ; fdh - SBC $xxxx, X
		dd      instfe		          ; feh - INC $xxxx, X
		dd      invalidInsByte            ; ffh - Invalid

; This is the flag table for all 255 bytes

flagTable       db      02h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h

		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h

		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h

		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h

		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h
		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h
		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h
		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h

		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h
		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h
		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h
		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h

		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h
		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h
		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h
		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h

		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h
		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h
		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h
		db      80h, 80h, 80h, 80h, 80h, 80h, 80h, 80h

; This is the flag table for all 255 bytes - no overflow or sign

flagNoOvTable   db      02h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h

		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h

		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h

		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h

		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h

		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h

		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h

		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h
		db      00h, 00h, 00h, 00h, 00h, 00h, 00h, 00h

_DATA           ends

; Support routines go here

;************************************************************************
;*                                                                      *
;* Execution points for instructions                                    *
;*                                                                      *
;************************************************************************

_TEXT           segment         dword public 'CODE'

		assume          CS:_TEXT

memoryread PROC
      push    edi
		push    esi ;push everything
		push    ebx
		push    ecx
		push    edx
		push    ebp

ifdef RESETPC
      denaturalizeESI
      mov m6502pc,si
endif

ifdef MSVC
 mov ecx,eax
endif
		call    [m6502MemoryRead]     ; Go call our handler!

		pop     ebp
		pop     edx
		pop     ecx
		pop     ebx
		pop     esi
      pop     edi
      ret
memoryread ENDP


memorywrite PROC
      push    edi
		push    ebx
		push    ecx
		push    edx
		push    esi
		push    ebp

ifdef RESETPC
      denaturalizeESI
      mov m6502pc,si
endif

ifdef MSVC
 mov ecx,eax
endif

		call    [m6502MemoryWrite]     ; Go call our handler!

		pop     ebp
		pop     esi
		pop     edx
		pop     ecx
		pop     ebx
      pop     edi

		test	al, al
		jz	noReloadPc

     ; Reload the program counter
      denaturalizeESI
		mov	edi, m6502rom	; Get our base address
      naturalizeESI
noReloadPc:
      xor eax,eax

      ret
memorywrite ENDP


;************************************************************************
;*                                                                      *
;* Main routines                                                        *
;*                                                                      *
;************************************************************************

; 00 - BRK

Alignment

inst00          proc

		Clocktime       7
      denaturalizeESI          ; Get our REAL PC
	  	add     si, 1           ; Add 2

;get stack segment
		mov	  al, byte ptr m6502s
      mov     ah,1           ;stack segment

;push PC
      dec     al
		mov     [eax + ebp], si ; Store our offset
		dec     al             ; Decrement!

;push flags
		mov     [eax + ebp], dh ; Store our flags
      or      byte ptr [eax + ebp],BREAK ;set break on stack
		dec     al
		mov     byte ptr m6502s, al ; Store it!

 ;get pc vector
		mov     si, [edi + 0fffeh] ; Place to get break address
      naturalizeESI

 ;twiddle flags
		or      dh, INTERRUPT       ; OR In our interupt flag
		and     dh, 0FFh-DECIMAL  ; remove out decimal flag

		ExecuteNextInstruction 7        ; Return to the main loop

inst00          endp

; 01 - ORA ($xx),X

Alignment

inst01          proc

		Clocktime       6

		FetchInstructionByte    ; Get our offset
;		mov     al, [ebp+eax]   ; Get the value it points to
;		add     al, bl          ; Add in X to it
;		mov     ax, [ebp+eax]   ; Go get the address
;		ReadMemoryByte          ; Go read it!
      add al,bl
		mov	ax, [ebp+eax]	; Get zero page address
      ReadMemoryByte

		or      dl, al          ; OR it in with the accumulator
		FlagSetAcc              ; Accumulator flag set
		ExecuteNextInstruction 6         ; Return to the main loop

inst01          endp

; 05 - ORA $xx

Alignment

inst05          proc

		Clocktime 3
		FetchInstructionByte    ; Get address to OR with
;		mov     al, [ebp+eax]   ; Zero page - go get it
;		or      dl, al          ; OR it in with the accumulator
      or     dl,[ebp+eax]
		FlagSetAcc              ; Set flags based upon accumulator
		ExecuteNextInstruction 3

inst05          endp

; 06 - ASL $xx

Alignment

inst06          proc

		Clocktime 5
		FetchInstructionByte    ; Get 0 page access address

      rcr dh,1
		shl   byte ptr [ebp+eax], 1           ; Left shift!
      rcl dh,1
		mov	al,[ebp+eax]	; Store our data back into zp
		FlagSet                 ; Set flags based on AL
		ExecuteNextInstruction 5

inst06          endp

; 08 - PHP

Alignment

inst08          proc

		Clocktime 3

		mov     al, m6502s     ; Get our stack pointer

		inc     ah              ; AH = 1 (stack page)

		mov     [ebp+eax], dh   ; Store flags
		dec     m6502s         ; Back our stack pointer down one!

		dec     ah              ; Zero AH
		ExecuteNextInstruction 3

inst08          endp

; 09 - ORA #xxh

Alignment

inst09          proc

		Clocktime 2
		FetchInstructionByte    ; Get value to and with A
		or      dl, al          ; And it in!
		FlagSetAcc              ; Set flags based on accumulator
		ExecuteNextInstruction 2

inst09          endp

; 0a - ASL A

Alignment

inst0a          proc

		Clocktime 2

;		and     dh, 0ffh - CARRY ; No carry!
     rcr  dh,1
		shl     dl, 1           ; Left shift!
     rcl dh,1
;		adc	dh, 0		; Set carry (if applicable)
		FlagSetAcc              ; Set flags based on accumulator
		ExecuteNextInstruction 2

inst0a          endp

; 0d - ORA $xxxx

Alignment

inst0d          proc

		Clocktime 4
		FetchInstructionWord    ; Get address to OR with
		ReadMemoryByte          ; Go read the data @ that address
		or      dl, al          ; OR it in with the accumulator
		FlagSetAcc              ; Set flags based upon accumulator
		ExecuteNextInstruction 4

inst0d          endp

; 0e - ASL $xxxx

Alignment

inst0e          proc

		Clocktime 6

		FetchInstructionWord    ; Get absolute address

 mov  bh,dl ;save acc
 push eax ;address to read
		ReadMemoryByte		; Go read our data

		and     dh, 0ffh - CARRY ; No carry!
      shl al,1
      adc dh,0    ;put carry back
		FlagSet

		mov	dl, al		; Byte to write in DL
 pop eax   ;address to write
		WriteMemoryByte		; Go write it!
 mov dl,bh
 xor bh,bh
		ExecuteNextInstruction 6

inst0e          endp

; 10 - BPL $xxxx

Alignment

inst10          proc

		Clocktime 2
		test    dh, SIGN        ; Are we positive?
		jz      doTakeRelativeJump ; Yup! Go do it!
		inc     esi             ; Skip over offset
		ExecuteNextInstruction 2

inst10          endp

; 11 - ORA ($xx), Y

Alignment

inst11          proc
		Clocktime 5

		FetchInstructionByte    ; Go get our address
		mov	ax, [ebp+eax]	; Get zero page address
		add	ax, cx		; Add in Y
      ReadMemoryByte
      or dl,al ;or it
		FlagSetAcc
		ExecuteNextInstruction 5

inst11          endp

; 15 - ORA $xx, X

Alignment

inst15          proc
                ;heynow
		Clocktime 4
		FetchInstructionByte    ; Get address to OR with
		add     al, bl          ; Add in X, too
		mov     al, [ebp+eax]   ; Go get it!
		or      dl, al          ; OR it in with the accumulator
		FlagSetAcc              ; Set flags based upon accumulator
		ExecuteNextInstruction 4

inst15          endp

; 16 - ASL $xx, X

Alignment

inst16          proc

		Clocktime 6

		FetchInstructionByte	; Go get our zero page
		add	al, bl		; Add X

      rcr dh,1
		shl   byte ptr [ebp+eax], 1           ; Left shift!
      rcl dh,1
      mov   al,[ebp+eax]
		FlagSet                 ; Set flags based on AL
		ExecuteNextInstruction 6

inst16          endp

; 18 - CLC

Alignment

inst18          proc

		Clocktime 2
		and     dh, 0ffh - CARRY ; No more carry!
		ExecuteNextInstruction 2

inst18          endp

; 19 - OR $xxxx, Y

Alignment

inst19          proc

		Clocktime 4
		FetchInstructionWord    ; Get address to AND with
		add     ax, cx          ; Add in Y
		ReadMemoryByte          ; Go read the data @ that address
		or      dl, al          ; OR it in with the accumulator
		FlagSetAcc              ; Set flags based upon accumulator
		ExecuteNextInstruction 4

inst19          endp

; 1d - OR $xxxx, X

Alignment

inst1d          proc

		Clocktime 5
		FetchInstructionWord    ; Get address to AND with
		add     ax, bx          ; Add in X
		ReadMemoryByte          ; Go read the data @ that address
		or      dl, al          ; OR it in with the accumulator
		FlagSetAcc              ; Set flags based upon accumulator
		ExecuteNextInstruction 5

inst1d          endp

; 1e - ASL $xxxx, X

Alignment

inst1e          proc

		Clocktime 7
		FetchInstructionWord    ; Get absolute address
		add	  ax, bx		; Add in X

 mov  bh,dl ;save acc
 push eax ;address to read
		ReadMemoryByte		; Go read our data

		;and     dh, 0ffh - CARRY ; No carry!
      rcr dh,1
      shl al,1
      rcl dh,1
      ;adc dh,0    ;put carry back
		FlagSet

		mov	dl, al		; Byte to write in DL
 pop eax   ;address to write
		WriteMemoryByte		; Go write it!
 mov dl,bh
 xor bh,bh
		ExecuteNextInstruction 7

inst1e          endp

; 20 - JSR $xxxx

Alignment

inst20          proc

		Clocktime 6

		FetchInstructionWord    ; Get our "to call" address
      push eax     ;heynow
    denaturalizeESI

ifdef LASTJSR
  mov m6502lastjsr,si ;store call addr
endif

		dec     si              ; Our offset to return to

		sub     m6502s, 2      ; Back up 2 bytes for stack push
		mov     ax, word ptr m6502s ; Our stack area

		xor     ah, ah          ; Store it!
		mov     [eax + ebp + 100h+1], si ; Store our return address

      pop esi    ;;get destination address of jump back
     naturalizeESI
		ExecuteNextInstruction 6         ; Return to our main fetch loop

inst20          endp

; 21 - AND ($xx), X

Alignment

inst21          proc

		Clocktime 6

		FetchInstructionByte    ; Go get our address
      add al,bl
		mov	ax, [ebp+eax]	; Get zero page address
      ReadMemoryByte
      and dl,al ;and it
		FlagSetAcc
		ExecuteNextInstruction 6

inst21          endp

; 24 - BIT $xx

Alignment

inst24          proc

		Clocktime 3

		FetchInstructionByte    ; Get address of data to do
		ReadMemoryByte          ; Go get the data

		mov     ah, al          ; A copy here, too
		and     ah, 0c0h        ; Bits 6 & 7
		and     al, dl          ; And AL with DL
		jnz     zeroFlag24

		or      dh, ZERO        ; Non-zero!
		jmp     short bitCont24

zeroFlag24:
		and     dh, 0ffh - ZERO ; We're Zero!

bitCont24:
		and     dh, 03fh        ; Knock out two upper bits
		or      dh, ah          ; OR In bits 6 & 7!
		ExecuteNextInstruction 3 ; Return to main loop!

inst24          endp

; 25 - AND $xx

Alignment

inst25          proc

		Clocktime 3      ;heynow
		FetchInstructionByte    ; Get address to OR with
		and     dl, [ebp+eax]   ; AND it in with the accumulator
		FlagSetAcc              ; Set flags based upon accumulator
		ExecuteNextInstruction 3

inst25          endp

; 26 - ROL $xx

Alignment

inst26          proc

		Clocktime 5

		FetchInstructionByte    ; Get 0 page access address
      rcr dh,1 ;get carry
		rcl byte ptr [ebp+eax], 1           ; Rotate byte
      rcl dh,1 ;put carry
      mov al,[ebp+eax]
		FlagSet
		ExecuteNextInstruction 5

inst26          endp

; 28 - PLP

Alignment

inst28          proc

		Clocktime 4

		inc     m6502s         ; Increment stack
		mov     al, m6502s     ; Get our stack address

;		inc     ah              ; Make it in the stack arena
		mov     dh, [ebp + eax+100h] ; Get our flag register
;		dec     ah              ; Zero AH again
		ExecuteNextInstruction 4

inst28          endp

; 29 - AND #xxh

Alignment

inst29          proc

		Clocktime 2
		FetchInstructionByte    ; Get value to and with A
		and     dl, al          ; And it in!
		FlagSetAcc              ; Set flags based on accumulator
		ExecuteNextInstruction 2

inst29          endp

; 2a - ROL A

Alignment

inst2a          proc

		Clocktime 2

      rcr dh,1
      rcl dl,1
      rcl dh,1

		FlagSetAcc
		ExecuteNextInstruction 2
inst2a          endp

; 2c - BIT $xxxx

Alignment

inst2c          proc

		Clocktime 4

		FetchInstructionWord    ; Get address of data to do
		ReadMemoryByte          ; Go get the data

		mov     ah, al          ; A copy here, too
		and     ah, 0c0h        ; Bits 6 & 7
		and     al, dl          ; And AL with DL
		jnz     zeroFlag

		or      dh, ZERO        ; Non-zero!
		jmp     short bitCont

zeroFlag:
		and     dh, 0ffh - ZERO ; We're Zero!

bitCont:
		and     dh, 03fh        ; Knock out two upper bits
		or      dh, ah          ; OR In bits 6 & 7!
		ExecuteNextInstruction 4 ; Return to main loop!

inst2c          endp

; 2D - AND $xxxx

Alignment

inst2d          proc

		Clocktime 4
		FetchInstructionWord    ; Go get our absolute address
		ReadMemoryByte          ; Get value to and in AL
		and     dl, al          ; And it with accumulator
		FlagSetAcc              ; Accumulator flag setting
		ExecuteNextInstruction 4

inst2d          endp

; 2e - ROL $xxxx

Alignment

inst2e          proc

		Clocktime 6

		FetchInstructionWord	; Get our address
 mov  bh,dl ;save acc
 push eax ;address to read
		ReadMemoryByte		; Go read our data

      rcr dh,1    ;get carry flag
      rcl al,1
      rcl dh,1    ;put carry back
		FlagSet

		mov	dl, al		; Byte to write in DL
 pop eax   ;address to write
		WriteMemoryByte		; Go write it!
 mov dl,bh
 xor bh,bh
		ExecuteNextInstruction 6
inst2e          endp

; 30 - BMI $xxxx

Alignment

inst30          proc

		Clocktime 2

		test    dh, SIGN        ; Signed?
		jnz     doTakeRelativeJump

		inc     esi             ; Don't take the jump
		ExecuteNextInstruction 2

inst30          endp

; 31 - AND ($xx), Y

Alignment

inst31          proc

		Clocktime 5

		FetchInstructionByte    ; Go get our address
		mov	ax, [ebp+eax]	; Get zero page address
		add	ax, cx		; Add in Y
      ReadMemoryByte
      and dl,al ;or it
		FlagSetAcc
		ExecuteNextInstruction 5

inst31          endp

; 35 - AND $xx, X

Alignment

inst35          proc

		Clocktime 4      ;heynow
		FetchInstructionByte    ; Get address to OR with
		add     al, bl          ; Add in X, too
		mov     al, [ebp+eax]   ; Go get it!
		and     dl, al          ; OR it in with the accumulator
		FlagSetAcc              ; Set flags based upon accumulator
		ExecuteNextInstruction 4

inst35          endp

; 36 - ROL $xx, X

Alignment

inst36          proc

		Clocktime 6
                 ;heynow
		FetchInstructionByte	; Go get our address
		add	al, bl		; Add X to it

      rcr     dh,1 ;;get carry flag
		rcl     byte ptr [ebp+eax], 1           ; Rotate AL!
      rcl     dh,1 ;;get carry back
   	mov	  al,[ebp+eax]	;
		FlagSet
		ExecuteNextInstruction 6

inst36          endp

; 38 - SEC

Alignment

inst38          proc

		Clocktime 2             ; Timing information

		or      dh, CARRY       ; Set carry!
		ExecuteNextInstruction 2 ; Return to the main loop!

inst38          endp

; 39 - AND $xxxx, Y

Alignment

inst39          proc

		Clocktime 4
		FetchInstructionWord    ; Get address to AND with
		add     ax, cx          ; Add in Y
		ReadMemoryByte          ; Go read the data @ that address
		and      dl, al         ; AND it in with the accumulator
		FlagSetAcc              ; Set flags based upon accumulator
		ExecuteNextInstruction 4

inst39          endp

; 3d - AND $xxxx, X

Alignment

inst3d          proc

		Clocktime 5
		FetchInstructionWord    ; Get address to AND with
		add     ax, bx          ; Add in X
		ReadMemoryByte          ; Go read the data @ that address
		and     dl, al          ; AND it in with the accumulator
		FlagSetAcc              ; Set flags based upon accumulator
		ExecuteNextInstruction 5

inst3d          endp

; 3e - ROL $xxxx, X

Alignment

inst3e          proc

		Clocktime 7

		FetchInstructionWord	; Get our address
		add	ax, bx		; Add in X
 mov  bh,dl ;save acc
 push eax
		ReadMemoryByte		; Go read our data
      rcr  dh,1 ;get carry
      rcl al,1 ; rotate al
      rcl dh,1 ;put carry back
		FlagSet

		mov	dl, al		; Byte to write in DL
 pop eax
		WriteMemoryByte		; Go write it!
 mov dl,bh
 xor bh,bh
		ExecuteNextInstruction 7

inst3e          endp

; 40 - RTI

Alignment

inst40          proc

		Clocktime 6

		xor     esi, esi        ; Zero ESI for later

		mov     ax, word ptr m6502s    ; Get our stack, man
		add     m6502s, 3      ; Pop off a word & a byte
		mov     ah, 1h          ; Stack "page"

  ;pop flags
		inc     eax             ; Go get it!
		mov     dh, [ebp + eax] ; Get our flags
      or      dh,RESERVED

  ;get return address
		inc     eax             ; And now our address to return to
		mov     si, [ebp + eax] ; Get address to return to (+ 1)
    naturalizeESI
		mov	byte ptr inNmi, 0 ; Make sure we're not in an NMI anymore
		ExecuteNextInstruction 6

inst40          endp

; 41 - EOR ($xx), X

Alignment

inst41          proc

		Clocktime 6

		FetchInstructionByte    ; Go get our address
      add al,bl
		mov	ax, [ebp+eax]	; Get zero page address
;		add	eax, ebx		; Add in X
      ReadMemoryByte
      xor dl,al ;or it
		FlagSetAcc
		ExecuteNextInstruction 6

inst41          endp

; 45 - EOR $xx

Alignment

inst45          proc

		Clocktime 3

		FetchInstructionByte    ; Get value to XOR with
		mov     al, [ebp+eax]   ; Go get our address
		xor     dl, al          ; XOR It!
		FlagSetAcc
		ExecuteNextInstruction 3

inst45          endp

; 46 - LSR $xx

Alignment

inst46          proc

		Clocktime 5
		FetchInstructionByte    ; Go get the new address to shift

      and   dh,07Fh-CARRY ;remove negative
		shr   byte ptr [ebp+eax], 1           ; Shift right!
      adc   dh,0    ;get carry back
      mov   al,[ebp+eax]
		FlagSet

		ExecuteNextInstruction 5
inst46          endp

; 48 - PHA

Alignment

inst48          proc

		Clocktime 3

		mov     al, m6502s     ; Get our stack pointer
		dec     m6502s         ; Back our stack pointer down one!

;		mov     ah, 01h         ; Stack "segment"
;      inc     ah
		mov     [ebp+eax+100h], dl   ; Store accumulator
;		dec     ah              ; Zero AH
		ExecuteNextInstruction 3

inst48          endp

; 49 - EOR xxh

Alignment

inst49          proc

		Clocktime 2

		FetchInstructionByte    ; Get value to XOR with
		xor     dl, al          ; XOR It!
		FlagSetAcc
		ExecuteNextInstruction 2

inst49          endp

; 4a - LSR A

Alignment

inst4a          proc

		Clocktime 2

		and   dh, 07fh - CARRY ; Get rid of carry
		shr	dl, 1		; LSR A!
		adc	dh, 0		; Set carry (if applicable)

		FlagSetNoOvAcc          ; Set the flags!
		ExecuteNextInstruction 2

inst4a          endp

; 4c - JMP $xxxx

Alignment

inst4c          proc
               ;heynow
		Clocktime 3
		FetchInstructionWord    ; Go get new address to jump to
		mov     esi, eax        ; Move in our new address
;;;		add     esi, edi        ; Add in the base offset
   naturalizeESI
		ExecuteNextInstruction 3         ; New address!

inst4c          endp

; 4d - EOR $xxxx

Alignment

inst4d		proc

		Clocktime 4

		FetchInstructionWord	; Get our data address
		ReadMemoryByte		; Go read it!
		xor	dl, al		; XOR It with A
		FlagSetAcc		; Set flags based upon accumulator
		ExecuteNextInstruction 4

inst4d		endp

; 4e - LSR $xxxx

Alignment

inst4e          proc
		Clocktime 6
		FetchInstructionWord    ; Go get the new address to shift

 mov  bh,dl ;save acc
 push eax ;address to read
		ReadMemoryByte		; Go read our data

		and     dh, 07fh - CARRY ; No carry!
      shr al,1
      adc dh,0    ;put carry back
		FlagSet

		mov	dl, al		; Byte to write in DL
 pop eax   ;address to write
		WriteMemoryByte		; Go write it!
 mov dl,bh
 xor bh,bh
		ExecuteNextInstruction 6

inst4e          endp

; 50 - BVC $xxxx

Alignment

inst50          proc

		Clocktime 2
		test    dh, OVERFLOW    ; Is overflow set?
		jz      doTakeRelativeJump ; Nope!! Jump!

		inc     esi             ; Skip relative offset
		ExecuteNextInstruction 2

inst50          endp

; 51 - EOR ($xx), Y

Alignment

inst51          proc

		Clocktime 5

		FetchInstructionByte    ; Go get our address
		mov	ax, [ebp+eax]	; Get zero page address
		add	ax, cx		; Add in Y
      ReadMemoryByte
      xor dl,al ;or it
		FlagSetAcc
		ExecuteNextInstruction 5

inst51          endp

; 55 - EOR $xx, X

Alignment

inst55          proc

		Clocktime 4

		FetchInstructionByte    ; Go get our address
		add     al, bl          ; Add in X, too
		xor     dl, byte ptr [ebp+eax]   ; XOR It!
		FlagSetAcc
		ExecuteNextInstruction 4

inst55          endp

; 56 - LSR $xx,X

Alignment

inst56          proc

		Clocktime 6
		FetchInstructionByte    ; Go get the new address to shift
		add	al, bl		; Add in X

      and   dh,07Fh-CARRY ;remove negative
		shr   byte ptr [ebp+eax], 1           ; Shift right!
      adc   dh,0    ;get carry back
      mov   al,[ebp+eax]  ;get back for flags
		FlagSet
		ExecuteNextInstruction 6
inst56          endp

; 58 - CLI

Alignment

inst58		proc

		Clocktime 2

		and	dh, 0ffh - INTERRUPT ; Allow interrupts again
		ExecuteNextInstruction 2

inst58		endp

; 59 - EOR $xxxx, Y

Alignment

inst59          proc

		Clocktime 4
		FetchInstructionWord    ; Get address to AND with
		add     ax, cx          ; Add in Y
		ReadMemoryByte          ; Go read the data @ that address
		xor      dl, al         ; XOR it in with the accumulator
		FlagSetAcc              ; Set flags based upon accumulator
		ExecuteNextInstruction 4

inst59          endp

; 5d - EOR $xxxx, X

Alignment

inst5d          proc

		Clocktime 5

		FetchInstructionWord    ; Go get our address
		add    ax, bx          ; Add in X
		ReadMemoryByte          ; Go get the byte
		xor     dl, al          ; XOR It!
		FlagSetAcc
		ExecuteNextInstruction 5

inst5d          endp

; 5e - LSR $xxxx, X

Alignment

inst5e          proc

		Clocktime 7
		FetchInstructionWord    ; Go get the new address to shift
		add	ax, bx		; Add in X
mov  bh,dl ;save acc
push eax
		ReadMemoryByte          ; Go read the memory byte

		and     dh, 07fh - CARRY ; Get rid of carry & negative
		shr     al, 1           ; Shift right!
      adc     dh,0    ;;get carry back
		FlagSet		; Need to BEFORE the restore

		mov     dl, al          ; Byte to write
pop eax
		WriteMemoryByte
mov dl,bh
xor bh,bh
		ExecuteNextInstruction 7

inst5e          endp

; 60 - RTS

Alignment

inst60          proc

		Clocktime 6

		xor     esi, esi        ; Zero ESI for later

		mov     ax, word ptr m6502s    ; Get our stack, man
		add     m6502s, 2      ; Pop off a word

		mov     ah, 1h          ; Stack "page"
 ;pop PC
   	inc     al
		mov     si, [ebp + eax] ; Get address to return to (+ 1)
		inc     si              ; To the next address
      naturalizeESI
		ExecuteNextInstruction 6

inst60          endp

; 61 - ADC $(xx,X)

Alignment

inst61          proc

		Clocktime 6
		FetchInstructionByte    ; Get our address
      add al,bl ;;add x
		mov     ax, [ebp + eax] ; Get our offset
		ReadMemoryByte          ; Go read a memory byte
		AdcMacro
		ExecuteNextInstruction 6

inst61          endp

; 65 - ADC $xx

Alignment

inst65          proc

		Clocktime 3

		FetchInstructionByte    ; Get address
		mov     al, [ebp + eax] ; Get byte to add to accumulator
		AdcMacro                ; Add with carry!
		ExecuteNextInstruction 3

inst65          endp

; 66 - ROR $xx

Alignment

inst66          proc

		Clocktime 5

		FetchInstructionByte    ; Get next byte

      rcr dh,1 ;get carry
		rcr byte ptr [ebp+eax], 1           ; Rotate byte
      rcl dh,1 ;put carry
      mov al,[ebp+eax]
		FlagSet
		ExecuteNextInstruction 5

inst66          endp

; 68 - PLA

Alignment

inst68          proc

		Clocktime 4

		inc     m6502s         ; Increment stack
		mov     al, m6502s     ; Get our stack address

;;;		inc     ah              ; Make it in the stack arena
		mov     dl, [ebp + eax+100h] ; Get our A register
		FlagSetAcc              ; Go set its flag
		ExecuteNextInstruction 4

inst68          endp

; 69 - ADC #$xx

Alignment

inst69          proc

		Clocktime 2
		FetchInstructionByte
		AdcMacro                ; Do ADC!
		ExecuteNextInstruction 2

inst69          endp

; 6a - ROR A

Alignment

inst6a          proc

		Clocktime 2

      rcr  dh,1 ;get carry
      rcr dl,1
      rcl dh,1 ;store carry
		FlagSetAcc
		ExecuteNextInstruction 2

inst6a          endp

; 6c - JMP ($xxxx)

Alignment

inst6c		proc
               ;heynow
		Clocktime 5
		FetchInstructionWord    ; Go get new address to jump to
;		mov	ax, [ebp+eax]	; Go get our new indirect address
;		mov     esi, eax        ; Move in our new address
;		add     esi, edi        ; Add in the base offset
;		xor     ah, ah          ; Zero AH for later
;		ExecuteNextInstruction 5 ; New address!

      xor esi,esi
      test ah,ah
      jns notrom
		mov	  si, [edi+eax]	; Go get our new indirect address
      jmp dojump
notrom:
      cmp ah,20h
      jae notlowerram
      and ah,7h
notlowerram:
      mov	  si, [ebp+eax]	; Go get our new indirect address
dojump:
      naturalizeESI
		ExecuteNextInstruction 5 ; New address!

inst6c		endp

; 6d - ADC $xxxx

Alignment

inst6d          proc

		Clocktime 4
		FetchInstructionWord
		ReadMemoryByte          ; Go read in value to add
		AdcMacro                ; Add with carry!
		ExecuteNextInstruction 4

inst6d          endp

; 6e - ROR $xxxxh

Alignment

inst6e          proc

		Clocktime 6

		FetchInstructionWord    ; Get next word

 mov  bh,dl ;save acc
 push eax ;address to read
		ReadMemoryByte		; Go read our data

      rcr dh,1    ;get carry flag
      rcr al,1    ;shift data
      rcl dh,1    ;put carry back
		FlagSet     ;set flags off al

		mov	dl, al		; Byte to write in DL
 pop eax   ;address to write
		WriteMemoryByte		; Go write it!
 mov dl,bh ;restore acc
 xor bh,bh
		ExecuteNextInstruction 6

inst6e          endp

; 70 - BVS $xxxx

Alignment

inst70          proc

		Clocktime 2
		test    dh, OVERFLOW    ; Is overflow set?
		jnz     doTakeRelativeJump ; Yup! Jump!

		inc     esi             ; Skip relative offset
		ExecuteNextInstruction 2

inst70          endp

; 71 - ADC $(xx,Y)

Alignment

inst71          proc

		Clocktime 5
		FetchInstructionByte    ; Get our address
		mov     ax, [ebp + eax] ; Get our offset
		add     ax, cx          ; Add Y
		ReadMemoryByte          ; Go read a memory byte
		AdcMacro
		ExecuteNextInstruction 5

inst71          endp

; 75 - ADC $xx, X

Alignment

inst75          proc

		Clocktime 4
		FetchInstructionByte
		add     al, bl          ; Add in X!
		mov     al, [ebp+eax]   ; Go get the data!
		AdcMacro                ; Add with carry!
		ExecuteNextInstruction 4

inst75          endp

; 76 - ROR $xx, X

Alignment

inst76          proc

		Clocktime 6

		FetchInstructionByte    ; Get next byte
		add	al, bl		; Add in X

      rcr dh,1 ;get carry
		rcr byte ptr [ebp+eax], 1           ; Rotate byte
      rcl dh,1 ;put carry
      mov al,[ebp+eax]
      FlagSet
   	ExecuteNextInstruction 6

inst76          endp

; 78 - SEI

Alignment

inst78		proc

		Clocktime 2

		or	dh, INTERRUPT	; Set interrupt disable
		ExecuteNextInstruction 2

inst78		endp

; 79 - ADC $xxxx,Y

Alignment

inst79          proc

		Clocktime 4
		FetchInstructionWord
		add     ax, cx          ; Add in Y!
		ReadMemoryByte          ; Go read in value to add
		AdcMacro                ; Add with carry!
		ExecuteNextInstruction 4

inst79          endp

; 7d - ADC $xxxx,X

Alignment

inst7d          proc

		Clocktime 5
		FetchInstructionWord
		add     ax, bx          ; Add in X!
		ReadMemoryByte          ; Go read in value to add
		AdcMacro                ; Add with carry!
		ExecuteNextInstruction 5

inst7d          endp

; 7e - ROR $xxxxh, X

Alignment

inst7e          proc

		Clocktime 7

		FetchInstructionWord    ; Get next word
		add	ax, bx		; Add in X

 mov  bh,dl ;save acc ;save acc
 push eax ;address to read
		ReadMemoryByte		; Go read our data

      rcr dh,1    ;get carry flag
      rcr al,1    ;shift data
      rcl dh,1    ;put carry back
		FlagSet     ;set flags off al

		mov	dl, al		; Byte to write in DL
 pop eax   ;address to write
		WriteMemoryByte		; Go write it!
 mov dl,bh ;restore acc
 xor bh,bh
		ExecuteNextInstruction 7

inst7e          endp

; 81 - STA $(xx,X)

Alignment

inst81          proc

		Clocktime 6
		FetchInstructionByte    ; Get our address
      add al,bl
		mov     ax, [ebp + eax] ; Get our offset
;;		add     eax, ebx          ; Add X
		WriteMemoryByte		; Go store A away
		ExecuteNextInstruction 6

inst81          endp

; 84 - STY $xx

Alignment

inst84          proc

		Clocktime 3
		FetchInstructionByte    ; Get instruction byte
		mov	[ebp+eax], cl	; Store Y in our zero page
		ExecuteNextInstruction 3

inst84          endp

; 85 - STA $xx

Alignment

inst85          proc

		Clocktime 3
		FetchInstructionByte    ; Get instruction byte
		mov	[ebp+eax], dl	; Store the accumulator (zero page)
		ExecuteNextInstruction 3

inst85          endp

; 86 - STX $56

Alignment

inst86          proc

		Clocktime 3
		FetchInstructionByte    ; Get our zero page address to write
		mov	[ebp+eax], bl	; Store X in zero page
		ExecuteNextInstruction 3

inst86          endp

; 88 - DEY

Alignment

inst88          proc

		Clocktime 2

		dec     cl              ; Decrement Y
		FlagSetY
		ExecuteNextInstruction 2

inst88          endp

; 8a - TXA

Alignment

inst8a          proc

		Clocktime 2
		mov     dl, bl          ; A = X
		FlagSetAcc              ; Go set flags from accumulator

		ExecuteNextInstruction 2

inst8a          endp

; 8c - STY $xxxx

Alignment

inst8c          proc

		Clocktime 4
		FetchInstructionWord    ; Get address to store
      push edx
      mov  dl,cl
	  ;	xchg    cl, dl          ; Swap for the sake of write
		WriteMemoryByte         ; Go write the data
	  ;	xchg    cl, dl          ; Swap for the sake of restoration
      pop edx
		ExecuteNextInstruction 4

inst8c          endp

; 8d - STA $xxxx

Alignment

inst8d          proc

		Clocktime 4
		FetchInstructionWord    ; Get address to store
		WriteMemoryByte         ; Go write the data
		ExecuteNextInstruction 4

inst8d          endp

; 8E - STX $xxxx

Alignment

inst8e          proc

		Clocktime 4
		FetchInstructionWord    ; Get address to store
;		xchg    bl, dl          ; Swap X & A
      push edx
      mov  dl,bl

		WriteMemoryByte         ; Go write the data
;		xchg    bl, dl          ; Swap them back
      pop edx
		ExecuteNextInstruction 4

inst8e          endp

; 90 - BCC $xxxx

Alignment

inst90          proc

		Clocktime 2
		test    dh, CARRY       ; Is carry set?
		jz      doTakeRelativeJump

		inc     esi             ; Skip relative offset
		ExecuteNextInstruction 2

inst90          endp

; 91 - STA ($xx), Y

Alignment

inst91          proc

		Clocktime 6

		FetchInstructionByte    ; Get our offset in AX (zero page)
		mov     ax, [ebp + eax] ; Get our address
		add     ax, cx          ; Put address in AX
		WriteMemoryByte         ; Go write the data
		ExecuteNextInstruction 6

inst91          endp

; 94 - STY $xx, X

Alignment

inst94          proc

		Clocktime 4

		FetchInstructionByte    ; Get our offset in AX (zero page)
		add     al, bl          ; Put address in AX
		mov	[ebp+eax], cl	; Store Y
		ExecuteNextInstruction 4

inst94          endp

; 95 - STA $xx, X

Alignment

inst95          proc

		Clocktime 4

		FetchInstructionByte    ; Get our offset in AX (zero page)
		add     al, bl          ; Put address in AX
		mov	[ebp+eax], dl	; Store A
		ExecuteNextInstruction 4

inst95          endp

; 96 - STX $xx, Y

Alignment

inst96          proc

		Clocktime 4

		FetchInstructionByte    ; Get our offset in AX (zero page)
		add     al, cl          ; Put address in AX
		mov	[ebp+eax], bl	; Store X
		ExecuteNextInstruction 4

inst96          endp

; 98 - TYA

Alignment

inst98          proc

		Clocktime 2
		mov     dl, cl          ; A = Y
		FlagSetAcc              ; Go set flags from accumulator

		ExecuteNextInstruction 2

inst98          endp

; 99 - STA $xxxx, Y

Alignment

inst99          proc

		Clocktime 5

		FetchInstructionWord
		add     ax, cx          ; Add Absolute addr to Y
		WriteMemoryByte         ; Go write accumulator!
		ExecuteNextInstruction 5

inst99          endp

; 9a - TXS

Alignment

inst9a          proc

		Clocktime 2

		mov     m6502s, bl     ; Store our new stack pointer

		ExecuteNextInstruction 2

inst9a          endp

; 9d - STA $xxxx, X

Alignment

inst9d          proc

		Clocktime 5

		FetchInstructionWord
		add     ax, bx          ; Add Absolute addr to X
		WriteMemoryByte         ; Go write accumulator!
		ExecuteNextInstruction 5

inst9d          endp

; a0 - LDY #xxh

Alignment

insta0          proc

		Clocktime 2

		FetchInstructionByte    ; Go get our immediate data
		mov     cl, al          ; Put it in Y
		FlagSet                 ; Set our flags according to AL
		ExecuteNextInstruction 2

insta0          endp

; a1 - LDA ($xx),X

Alignment

insta1          proc

		Clocktime 6
		FetchInstructionByte    ; Get our address
      add al,bl
		mov     ax, [ebp + eax] ; Get our offset
		ReadMemoryByte          ; Go read a memory byte
		mov     dl, al          ; Go read it in!
		FlagSetAcc              ; Flags based on accumulator
		ExecuteNextInstruction 6

insta1          endp

; a2 - LDX #xxh

Alignment

insta2          proc

		Clocktime 2

		FetchInstructionByte    ; Go get our immediate data
		mov     bl, al          ; Put it in X
		FlagSet                 ; Set our flags according to AL
		ExecuteNextInstruction 2

insta2          endp

; a4 - LDY $xx

Alignment

insta4          proc

		Clocktime 3

		FetchInstructionByte    ; Get zero page address
		mov	cl, byte ptr [ebp+eax] ; Go get our new Y
		FlagSetY                ; Set flags based on the Y register
		ExecuteNextInstruction 3

insta4          endp

; a5 - LDA $xx

Alignment

insta5          proc

		Clocktime 3

		FetchInstructionByte    ; Get zero page address
		mov	dl, byte ptr [ebp+eax] ; Go get our new accumulator
		FlagSetAcc              ; Set flags based on accumulator
		ExecuteNextInstruction 3

insta5          endp

; a6 - LDX $xx

Alignment

insta6          proc

		Clocktime 3

		FetchInstructionByte    ; Go get zero page address
		mov     bl, byte ptr [ebp + eax] ; Go get our data
		FlagSetX
		ExecuteNextInstruction 3

insta6          endp


; a8 - TAY

Alignment

insta8          proc

		Clocktime 2

		mov     cl, dl          ; Y=A now
		FlagSetAcc
		ExecuteNextInstruction 2

insta8          endp

; a9 - LDA #xx

Alignment

insta9          proc

		Clocktime 2

		FetchInstructionByte
		mov     dl, al          ; Put it in DL!
		FlagSet                 ; Go set our flags
		ExecuteNextInstruction 2

insta9          endp

; aa - TAX

Alignment

instaa          proc

		Clocktime 2

		mov     bl, dl          ; X = A
		FlagSetAcc
		ExecuteNextInstruction 2

instaa          endp

; ac - LDY $xxxx

Alignment

instac          proc

		Clocktime 4

		FetchInstructionWord    ; Get the address
		ReadMemoryByte          ; Go get byte at address
		mov     cl, al          ; Y = gotten byte
		FlagSetY                ; Go set flags based on Y
		ExecuteNextInstruction 4

instac          endp

; ad - LDA $xxxx

Alignment

instad          proc

		Clocktime 4

		FetchInstructionWord    ; Get the address
		ReadMemoryByte          ; Go get byte at address
		mov     dl, al          ; Get it in DL for accumulator load
		FlagSetAcc              ; Go set flags based on A
		ExecuteNextInstruction 4

instad          endp

; ae - LDX $xxxx

Alignment

instae          proc

		Clocktime 4

		FetchInstructionWord    ; Get the address
		ReadMemoryByte          ; Go get byte at address
		mov     bl, al          ; X = gotten byte
		FlagSetX                ; Go set flags based on X
		ExecuteNextInstruction 4

instae          endp

; b0 - BCS $xxxx

Alignment

instb0          proc

		Clocktime 2
		test    dh, CARRY       ; Is carry set?
		jnz     doTakeRelativeJump

		inc     esi             ; Skip relative offset
		ExecuteNextInstruction 2

instb0          endp

; b1 - LDA ($xx), Y

Alignment

instb1          proc

		Clocktime 5
		FetchInstructionByte    ; Get our address
		mov     ax, [ebp + eax] ; Get our offset
		add     ax, cx          ; Add Y
		ReadMemoryByte          ; Go read a memory byte
		mov     dl, al          ; Put it in the accumulator
		FlagSetAcc              ; Set the flags
		ExecuteNextInstruction 5

instb1          endp

; b4 - LDY $xx, X

Alignment

instb4          proc

		Clocktime 4

		FetchInstructionByte    ; Get the address
		add     al, bl          ; Add in the value of X
		mov     cl, [ebp + eax] ; Get our 16 bit address
		FlagSetY                ; Go set flags based on Y
		ExecuteNextInstruction 4

instb4          endp

; b5 - LDA $xx, X

Alignment

instb5          proc

		Clocktime 4

		FetchInstructionByte    ; Get the address
		add     al, bl          ; Add in the value of X
		mov     dl, [ebp + eax] ; Get our 16 bit address
		FlagSetAcc              ; Go set flags based on A
		ExecuteNextInstruction 4

instb5          endp

; b6 - LDX $xx, Y

Alignment

instb6		proc

		Clocktime 4

		FetchInstructionByte    ; Get the address
		add     al, cl          ; Add in the value of Y
		mov     bl, [ebp + eax] ; Get our 16 bit address
		FlagSetX                ; Go set flags based on X
		ExecuteNextInstruction 4

instb6		endp

; b8 - CLV

Alignment

instb8		proc

		Clocktime 2

		and	dh, 0ffh - OVERFLOW ; Get rid of overflow
		ExecuteNextInstruction 2

instb8		endp


; b9 - LDA $xxxx, Y

Alignment

instb9          proc

		Clocktime 4

		FetchInstructionWord    ; Get the address
		add     ax, cx          ; Add in the value of Y
		ReadMemoryByte          ; Go get byte at address
		mov     dl, al          ; Get it in DL for accumulator load
		FlagSetAcc              ; Go set flags based on A
		ExecuteNextInstruction 4

instb9          endp

; ba - TSX

Alignment

instba		proc

		Clocktime 2

		mov	bl, m6502s	; Get our S register

		FlagSetX
		ExecuteNextInstruction 2

instba		endp

; bc - LDY $xxxx, X

Alignment

instbc          proc

		Clocktime 4

		FetchInstructionWord    ; Get the address
		add     ax, bx          ; Add in the value of X
		ReadMemoryByte          ; Go get byte at address
		mov     cl, al          ; Get it in CL for Y
		FlagSet	                ; Go set flags based on A
		ExecuteNextInstruction 4

instbc          endp

; bd - LDA $xxxx, X

Alignment

instbd          proc

		Clocktime 4

		FetchInstructionWord    ; Get the address
		add     ax, bx          ; Add in the value of X
		ReadMemoryByte          ; Go get byte at address
		mov     dl, al          ; Get it in DL for accumulator load
		FlagSetAcc              ; Go set flags based on A
		ExecuteNextInstruction 4

instbd          endp

; be - LDX $xxxx, Y

Alignment

instbe          proc

		Clocktime 4

		FetchInstructionWord    ; Get the address
		add     ax, cx          ; Add in the value of Y
		ReadMemoryByte          ; Go get byte at address
		mov     bl, al          ; Get it in BL for X load
		FlagSet                 ; Go set flags based on A
		ExecuteNextInstruction 4

instbe          endp

; c0 - CPY #xxh

Alignment

instc0          proc

		Clocktime 2
		FetchInstructionByte    ; Get byte to subtract
		mov     ah, cl          ; Get a copy of the Y register
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarryc0
		or      dh, CARRY       ; Set carry!
noSetCarryc0:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet

;      and    dh,0FFh-CARRY
;      sub    al,cl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor  dh,SIGN+CARRY
		ExecuteNextInstruction 2

instc0          endp

; c1 - CMP ($xx), X

Alignment

instc1          proc

		Clocktime 6
		FetchInstructionByte	; Go get our zero page address
      add al,bl
		mov	ax, [ebp+eax]	; Go get our offset
		ReadMemoryByte		; Go read the data into AL
		mov     ah, dl          ; Get a copy of the accumulator
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarryc1
		or      dh, CARRY       ; Set carry!
noSetCarryc1:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet

;      and    dh,0FFh-CARRY
;      sub    al,dl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY
		ExecuteNextInstruction 6

instc1          endp

; c4 - CPY $xx

Alignment

instc4          proc

		Clocktime 3
		FetchInstructionByte    ; Get byte to subtract
		mov     al, [ebp+eax]   ; Go get the byte!
		mov     ah, cl          ; Get a copy of the Y register
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarryc4
		or      dh, CARRY       ; Set carry!
noSetCarryc4:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet
;      and    dh,0FFh-CARRY
;      sub    al,cl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY

		ExecuteNextInstruction 3

instc4          endp

; c5 - CMP $xxh

Alignment

instc5          proc

		Clocktime 3
		FetchInstructionByte    ; Get byte to subtract
		mov     al, [ebp+eax]   ; Go get our byte (in zp)
		mov     ah, dl          ; Get a copy of the accumulator
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarryc5
		or      dh, CARRY       ; Set carry!
noSetCarryc5:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet

;      and    dh,0FFh-CARRY
;      sub    al,dl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY

		ExecuteNextInstruction 3

instc5          endp

; c6 - DEC $xx

Alignment

instc6          proc

		Clocktime 5

		FetchInstructionByte    ; Go get address to deal with
		dec	byte ptr [ebp+eax] ; Decrement the byte
		mov	al, [ebp+eax]	; Go get the data
		FlagSet                 ; Set the flags
		ExecuteNextInstruction 5

instc6          endp

; c8 - INY

Alignment

instc8          proc

		Clocktime 2
		inc     cl              ; Increment Y
		FlagSetY                ; Set flags according to Y
		ExecuteNextInstruction 2         ; Return for the next instruction

instc8          endp

; c9 - CMP #xxh

Alignment

instc9          proc

		Clocktime 2
		FetchInstructionByte    ; Get byte to subtract
		mov     ah, dl          ; Get a copy of the accumulator
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarryc9
		or      dh, CARRY       ; Set carry!
noSetCarryc9:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet

;      and    dh,0FFh-CARRY
;      sub    al,dl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY
		ExecuteNextInstruction 2

instc9          endp

; ca - DEX

Alignment

instca          proc

		Clocktime 2

		dec     bl              ; Decrement X
		FlagSetX
		ExecuteNextInstruction 2

instca          endp

; cc - CPY $xxxx

Alignment

instcc          proc

		Clocktime 4
		FetchInstructionWord    ; Get word to subtract
		ReadMemoryByte
		mov     ah, cl          ; Get a copy of Y
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarrycc
		or      dh, CARRY       ; Set carry!
noSetCarrycc:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet


;      and    dh,0FFh-CARRY
;      sub    al,cl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY

		ExecuteNextInstruction 4

instcc          endp

; cd - CMP $xxxx

Alignment

instcd          proc

		Clocktime 4
		FetchInstructionWord    ; Get word to subtract
		ReadMemoryByte
		mov     ah, dl          ; Get a copy of the accumulator
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarrycd
		or      dh, CARRY       ; Set carry!
noSetCarrycd:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet
;      and    dh,0FFh-CARRY
;      sub    al,dl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY

		ExecuteNextInstruction 4

instcd          endp

; ce - DEC $xxxx

Alignment

instce          proc

		Clocktime 6

		FetchInstructionWord    ; Go get address to deal with
 mov  bh,dl ;save acc
 push eax
		ReadMemoryByte          ; Go get the byte to increment
		dec     al              ; Decrement our byte
		FlagSet                 ; Go set the flags
		mov     dl,  al         ; Put it in DL for the memory byte write
 pop eax
		WriteMemoryByte         ; Write it back
 mov dl,bh
 xor bh,bh
		ExecuteNextInstruction 6




;		FetchInstructionWord    ; Go get address to deal with
; push edx
; push eax
;		ReadMemoryByte          ; Go get the byte to increment
;      mov   dl,al
;pop  eax
;		dec     dl              ; Decrement our byte
;		WriteMemoryByte         ; Write it back
; FlagSetAcc
; pop edx
;		ExecuteNextInstruction 6


instce          endp

; d0 - BNE $xxxx

Alignment

instd0          proc

		Clocktime 2
		test    dh, ZERO        ; Not equal?
		jz      doTakeRelativeJump

		inc     esi             ; Don't take the jump
		ExecuteNextInstruction 2

doTakeRelativeJump:
		Clocktime 2		; Plus one for a taken jump
		FetchInstructionByte    ; Fetch it!
    denaturalizeESI
      cbw
		add     si, ax          ; Add in our offset
    naturalizeESI
		ExecuteNextInstruction 4

instd0          endp

; d1 - CMP ($xx), Y

Alignment

instd1          proc

		Clocktime 5
		FetchInstructionByte	; Go get our zero page address
		mov	ax, [ebp+eax]	; Go get our offset
		add	ax, cx		; Add in Y
		ReadMemoryByte		; Go read the data into AL
		mov     ah, dl          ; Get a copy of the accumulator
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarryd1
		or      dh, CARRY       ; Set carry!
noSetCarryd1:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet

;      and    dh,0FFh-CARRY
;      sub    al,dl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY
		ExecuteNextInstruction 5

instd1          endp

; d5 - CMP $xx, X

Alignment

instd5          proc

		Clocktime 4
		FetchInstructionByte    ; Get byte to subtract
		add     al, bl          ; Add in X
		mov     al, [ebp+eax]   ; Go get our byte (in zp)
		mov     ah, dl          ; Get a copy of the accumulator
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarryd5
		or      dh, CARRY       ; Set carry!
noSetCarryd5:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet
;      and    dh,0FFh-CARRY
;      sub    al,dl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY

		ExecuteNextInstruction 4

instd5          endp

; d6 - DEC $xx, X

Alignment

instd6          proc

		Clocktime 6
		FetchInstructionByte    ; Get offset within zero page
		add     al, bl          ; Add in the value of X
		dec     byte ptr [ebp + eax]    ; Decrement it!
		mov     al, [ebp + eax] ; Get the value
		FlagSet                 ; Go set the flags
		ExecuteNextInstruction  6 ; Return to the main loop

instd6          endp

; d8 - CLD

Alignment

instd8          proc

		Clocktime 2

		and     dh, 0ffh - DECIMAL ; Knock out decimal flag
		ExecuteNextInstruction 2

instd8          endp

; d9 - CMP $xxxx, Y

Alignment

instd9          proc

     		Clocktime 4
		FetchInstructionWord	; Go get our address
		add	ax, cx		; Add in Y
		ReadMemoryByte		; Go read the data into AL
		mov     ah, dl          ; Get a copy of the accumulator
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarryd9
		or      dh, CARRY       ; Set carry!
noSetCarryd9:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet
;      and    dh,0FFh-CARRY
;      sub    al,dl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY

		ExecuteNextInstruction 4

instd9          endp

; dd - CMP $xxxx, X

Alignment

instdd          proc

		Clocktime 5
		FetchInstructionWord	; Go get our address
		add	ax, bx		; Add in X
		ReadMemoryByte		; Go read the data into AL
		mov     ah, dl          ; Get a copy of the accumulator
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarrydd
		or      dh, CARRY       ; Set carry!
noSetCarrydd:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet
;      and    dh,0FFh-CARRY
;      sub    al,dl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY

		ExecuteNextInstruction 5

instdd          endp

; de - DEC $xxxx, X

Alignment

instde          proc

		Clocktime 7
		FetchInstructionWord    ; Go get address to deal with
		add     ax, bx          ; Add in X
 mov  bh,dl ;save acc
 push eax
		ReadMemoryByte          ; Go get the byte to increment
		dec     al              ; Decrement our byte
		FlagSet                 ; Go set the flags
		mov     dl,  al         ; Put it in DL for the memory byte write
 pop eax
		WriteMemoryByte         ; Write it back
 mov dl,bh
 xor bh,bh
		ExecuteNextInstruction 7

instde          endp

; e0 - CPX #xxh

Alignment

inste0          proc

		Clocktime 2
		FetchInstructionByte    ; Get byte to subtract
		mov     ah, bl          ; Get a copy of the X register
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarrye0
		or      dh, CARRY       ; Set carry!
noSetCarrye0:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet
;      and    dh,0FFh-CARRY
;      sub    al,bl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY

		ExecuteNextInstruction 2

inste0          endp

; e1 - SBC $(xx,X)

Alignment

inste1          proc

		Clocktime 6
		FetchInstructionByte    ; Get our address
      add    al,bl ;add x
		mov     ax, [ebp + eax] ; Get our offset
;;;;		add     eax, ebx          ; Add X
		ReadMemoryByte          ; Go read a memory byte
		SbcMacro
		ExecuteNextInstruction 6

inste1          endp

; e4 - CPX $xx

Alignment

inste4          proc

		Clocktime 3
		FetchInstructionByte    ; Get byte to subtract
		mov     al, [ebp+eax]   ; Go get the data
		mov     ah, bl          ; Get a copy of the X register
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarrye4
		or      dh, CARRY       ; Set carry!
noSetCarrye4:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet
;      and    dh,0FFh-CARRY
;      sub    al,bl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY

		ExecuteNextInstruction 3

inste4          endp

; e5 - SBC $xx

Alignment

inste5          proc

		Clocktime 3
		FetchInstructionByte    ; Get zero page address
		mov     al, [ebp+eax]   ; Go get the value to subtract
		SbcMacro                ; Do the subtract
		ExecuteNextInstruction 3

inste5          endp

; e6 - INC $xx

Alignment

inste6          proc

		Clocktime 5

		FetchInstructionByte    ; Go get address to deal with
		inc	byte ptr [ebp+eax]
		mov	al, [ebp+eax]
		FlagSet                 ; Set the flags
		ExecuteNextInstruction 5

inste6          endp

; e8 - INX

Alignment

inste8          proc

		Clocktime 2

		inc     bl              ; Increment X
		FlagSetX
		ExecuteNextInstruction 2

inste8          endp

; e9 - SBC #$xx

Alignment

inste9          proc

		Clocktime 2

		FetchInstructionByte    ; Get byte to subtract
		SbcMacro                ; Do SBC
		ExecuteNextInstruction 2

inste9          endp

; ea - NOP

Alignment

instea		proc

		Clocktime 2
		ExecuteNextInstruction 2	; Do nothing!

instea		endp

; ec - CPX $xxxx

Alignment

instec          proc

		Clocktime 4
		FetchInstructionWord    ; Get byte to subtract
		ReadMemoryByte		; Go read the data
		mov     ah, bl          ; Get a copy of the X register
		and     dh, 0ffh - CARRY ; No carry!
		sub     ah, al          ; Subtract!
		jc      noSetCarryec
		or      dh, CARRY       ; Set carry!
noSetCarryec:
		mov     al, ah          ; Get our value in AL for compare
		FlagSet

;      and    dh,0FFh-CARRY
;      sub    al,bl          ; Subtract!
;      adc    dh,0
;		FlagSet
;      xor     dh,SIGN+CARRY

		ExecuteNextInstruction 4

instec          endp

; ed - SBC $xxxx

Alignment

insted          proc

		Clocktime 4

		FetchInstructionWord    ; Get the address
		ReadMemoryByte          ; Go get byte at address
		SbcMacro                ; Subtract with carry
		ExecuteNextInstruction 4

insted          endp

; ee - INC $xxxx

Alignment

instee          proc

		Clocktime 6

		FetchInstructionWord    ; Go get address to deal with

 mov  bh,dl ;save acc
 push eax
		ReadMemoryByte          ; Go get the byte to increment
		inc     al              ; increment our byte
		FlagSet                 ; Go set the flags
		mov     dl,  al         ; Put it in DL for the memory byte write
 pop eax
		WriteMemoryByte         ; Write it back
 mov dl,bh
 xor bh,bh
		ExecuteNextInstruction 6

instee          endp

; f0 - BEQ $xxxx

Alignment

instf0          proc

		Clocktime 2
		test    dh, ZERO        ; Not equal?
		jnz     doTakeRelativeJump

		inc     esi             ; Don't take the jump
		ExecuteNextInstruction 2

instf0          endp

; f1 - SBC $(xx,Y)

Alignment

instf1          proc

		Clocktime 5
		FetchInstructionByte    ; Get our address
		mov     ax, [ebp + eax] ; Get our offset
		add     ax, cx          ; Add Y
		ReadMemoryByte          ; Go read a memory byte
		SbcMacro
		ExecuteNextInstruction 5

instf1          endp

; f5 - SBC $xx, X

Alignment

instf5          proc

		Clocktime 4

		FetchInstructionByte    ; Get our zero page address
		add     al, bl          ; Add in X
		mov     al, [ebp + eax] ; Go get it!
		SbcMacro                ; Subtract!
		ExecuteNextInstruction 4

instf5          endp

; f6 - INC $xx, X

Alignment

instf6          proc

		Clocktime 6

		FetchInstructionByte    ; Get our zero page address
		add     al, bl          ; Add in X
		inc     byte ptr [ebp + eax] ; Increment it!
		mov     al, [ebp + eax] ; Go get the incremented value
		FlagSet                 ; Go set flags
		ExecuteNextInstruction 6 ; Return to the main loop

instf6          endp

; f8 - SED

Alignment

instf8          proc

		Clocktime 2

		or      dh, DECIMAL     ; Set Decimal mode
		ExecuteNextInstruction 2

instf8          endp

; f9 - SBC $xxxx, Y

Alignment

instf9          proc

		Clocktime 4

		FetchInstructionWord    ; Get the address
		add     ax, cx          ; Add in the value of Y
		ReadMemoryByte          ; Go get byte at address
		SbcMacro                ; Subtract with carry
		ExecuteNextInstruction 4

instf9          endp

; fd - SBC $xxxx, X

Alignment

instfd          proc

		Clocktime 5

		FetchInstructionWord    ; Get the address
		add     ax, bx          ; Add in the value of X
		ReadMemoryByte          ; Go get byte at address
		SbcMacro                ; Subtract with carry
		ExecuteNextInstruction 5

instfd          endp


; fe - INC $xxxx, X

Alignment

instfe          proc

		Clocktime 7

		FetchInstructionWord    ; Go get address to deal with
		add	ax, bx		; Add in X
 mov  bh,dl ;save acc
 push eax
		ReadMemoryByte          ; Go get the byte to increment
		inc     al              ; increment our byte
		FlagSet                 ; Go set the flags
		mov     dl,  al         ; Put it in DL for the memory byte write
 pop eax
		WriteMemoryByte         ; Write it back
 mov dl,bh
 xor bh,bh
		ExecuteNextInstruction 7

instfe          endp

; Support routines go here

;************************************************************************
;                                                                       *
; Name  : m6502reset()                                                  *
;                                                                       *
; Entry : Nothing                                                       *
;                                                                       *
; Exit  : Nothing                                                       *
;                                                                       *
; Description:                                                          *
;                                                                       *
; This routine will reset the 6502's flags, registers, & interrupt modes *
;                                                                       *
;************************************************************************

Alignment

m6502reset	proc
			push    ebp

		xor     eax, eax        ; Zero AX

		mov     ebp, m6502ram         ; Get the base address
		mov	inNmi, al	; We're not in an NMI
		mov     m6502x, al
		mov     m6502y, al
		mov     m6502s, 0ffh
		mov     m6502clockticks, eax

		mov     ah, RESERVED + ZERO         ; Reset value for 6502
		mov     m6502af, ax    ; Store it and the accumulator

		mov     ebp, m6502rom         ; Get the base address
		mov     ax, 0fffch      ; Reset vector
		mov     ax, [ebp + eax] ; Get our reset vector address

		mov     m6502pc, ax    ; Store our new PC

		pop     ebp

		ret

m6502reset	endp

;************************************************************************
;                                                                       *
; Name  : m6502nmi_                                                     *
;                                                                       *
; Entry : Nothing                                                       *
;                                                                       *
; Exit  : 0=NMI was taken. NMI Is always taken                          *
;                                                                       *
; Description:                                                          *
;                                                                       *
; This routine simulates a virtual NMI to the emulated 6502. No code    *
; execution actually takes place - only the PC is reset and the current *
; PC gets pushed to the stack.                                          *
;                                                                       *
;************************************************************************

Alignment

m6502nmi	proc

		xor	eax, eax		; Zero EAX

		inc	inNmi

		push    edi
		push    ebx

		Clocktime 7 			; We eat up 7 clock ticks


		mov     edi, [m6502ram]       ; Get pointer to game image
		mov     ebx, 0100h              ; Point to stack page
		mov     bl, [m6502s]           ; Get our S reg pointer

;push pc
		mov     ax, [m6502pc]          ; Get our PC
		mov     [edi + ebx], ah         ; Store it!
		dec     bl
		mov     [edi + ebx], al         ; Store it!
		dec     bl

;push flags
		mov     al, byte ptr [m6502af+1] ; Get our flags
		and     al, 0FFh-BREAK         ; erase breakflag on stack
      mov     [edi + ebx], al         ; Store flags
		dec     bl
		mov     m6502s, bl             ; Store S reg pointer

;change pc to nmi vector
      mov     edi,[m6502rom]
		mov     ax, [edi+0fffah]        ; Get our start vector!
		mov     m6502pc, ax            ; Store our new PC

;twiddle flags
      and     byte ptr m6502af + 1,0FFh-DECIMAL    ;turn off decimal mode
		or      m6502af + 1, RESERVED  ; Turn on something, dont disable interrupt

		pop     ebx
		pop     edi

		mov     eax, -1			; Indicate that we took the NMI
		ret
m6502nmi	endp


;************************************************************************
;                                                                       *
; Name  : m6502int_                                                     *
;                                                                       *
; Entry : Nothing                                                       *
;                                                                       *
; Exit  : 0=IRQ was taken, Non-zero=already in interrupt                *
;                                                                       *
; Description:                                                          *
;                                                                       *
; This routine simulates a virtual IRQ to the emulated 6502. No code    *
; execution actually takes place - only the PC is reset and the current *
; PC gets pushed to the stack.                                          *
;                                                                       *
;************************************************************************

Alignment

m6502int	proc
		xor	eax, eax		; Zero our result code

		test	byte ptr m6502af + 1, INTERRUPT	; Are interrupts disabled?
		jnz	notTakenint

		Clocktime 7 			; We eat up 7 clock ticks

		push    edi
		push    ebx

		mov     ebx, 0100h              ; Point to stack page

		mov     edi, [m6502ram]       ; Get pointer to game image
		mov     bl, [m6502s]           ; Get our S reg pointer

     ;push pc
		mov     ax, [m6502pc]          ; Get our PC
		mov     [edi + ebx], ah         ; Store it!
		dec     bl
		mov     [edi + ebx], al         ; Store it!
		dec     bl

     ;push flags with BREAK clear
		mov     al, byte ptr [m6502af+1] ; Get our flags
		and     al, 0FFh-BREAK         ; erase breakflag on stack
		mov     [edi + ebx], al         ; Store flags
		dec     bl
		mov     m6502s, bl             ; Store S reg pointer

     ;change pc to int vector
      mov     edi,[m6502rom]
		mov     ax, [edi+0fffeh]        ; Get our start vector!
		mov     m6502pc, ax            ; Store our new PC

     ;adjust flags
      and     byte ptr m6502af + 1,0FFh-DECIMAL    ;turn off decimal mode
		or	     byte ptr m6502af + 1, RESERVED+INTERRUPT  ;turn on interrupt flag

		pop     ebx
		pop     edi

		mov     eax, -1			; Indicate IRQ was taken

notTakenint:
		ret

m6502int	endp


;************************************************************************
;                                                                       *
; Name  : m6502GetContextSize                                           *
;                                                                       *
; Entry : Nothing                                                       *
;                                                                       *
; Exit  : Size of context array (in bytes)                              *
;                                                                       *
; Description:                                                          *
;                                                                       *
; This routine returns the # of bytes needed for a context switch for   *
; multi-CPU operation.                                                  *
;                                                                       *
;************************************************************************

ifdef MULTI_CPU

Alignment

m6502GetContextSize	proc

		mov     eax, ((offset ContextEnd - offset ContextBegin) AND 0FFFFFFFCh) + 4 ; DWORD Align it!
		ret

m6502GetContextSize	endp

endif

;************************************************************************
;                                                                       *
; Name  : m6502SetContext                                               *
;                                                                       *
; Entry : Address of context in EAX                                     *
;                                                                       *
; Exit  : Nothing                                                       *
;                                                                       *
; Description:                                                          *
;                                                                       *
; This routine will set a context. On input, a source address for the   *
; context is given.                                                     *
;                                                                       *
;************************************************************************

ifdef MULTI_CPU

Alignment

m6502SetContext	proc	contextAddr:DWORD

		mov	eax, contextAddr

		push    esi             ; Save registers we use
		push    edi
		push    ecx

		mov     ecx, (((offset ContextEnd - offset ContextBegin) AND 0fffffffch)) SHR 2
		mov     esi, eax        ; Source address in ESI
		mov     edi, offset ContextBegin
		rep     movsd           ; Move it as fast as we can!

		pop     ecx
		pop     edi
		pop     esi

		ret                     ; No return code

m6502SetContext	endp

endif	; MULTI_CPU

;************************************************************************
;                                                                       *
; Name  : m6502GetContext                                               *
;                                                                       *
; Entry : Address of context in EAX                                     *
;                                                                       *
; Exit  : Nothing                                                       *
;                                                                       *
; Description:                                                          *
;                                                                       *
; This routine will get a context. On input, a dest. address for the    *
; context is given.                                                     *
;                                                                       *
;************************************************************************

ifdef MULTI_CPU

Alignment

m6502GetContext	proc

		push    esi             ; Save registers we use
		push    edi
		push    ecx

		mov     edi, eax        ; Destination address in EDI

		mov     ecx, (((offset ContextEnd - offset ContextBegin) AND 0fffffffch)) SHR 2
		mov     esi, offset ContextBegin
		rep     movsd           ; Move it as fast as we can!

		pop     ecx
		pop     edi
		pop     esi

		ret                     ; No return code
m6502GetContext		endp

endif ; MULTI_CPU

;************************************************************************
;                                                                       *
; Name  : m6502GetInfo                                                  *
;                                                                       *
; Entry : Nothing                                                       *
;                                                                       *
; Exit  : EAX Contains bitwise information:                             *
;                                                                       *
; Bits      31-22           21  20  19  18  17  16  15-8   7-0          *
;          Unused           |   |   |   |   |   |   Major  Minor        *
;                           |   |   |   |   |   |   Rev    Rev          *
;                           |   |   |   |   |   |                       *
;                           |   |   |   |   |   +-0=UniCPU, 1=MultiCPU  *
;                           |   |   |   |   |                           *
;                           |   |   |   |   +-1=Timing info available   *
;                           |   |   |   |                               *
;                           |   |   |   +-1=SYNC emulation enabled      *
;                           |   |   |                                   *
;                           |   |   +-1=Bounds checking enabled         *
;                           |   |                                       *
;                           |   +-1=Instr. count, 0=Cycle count	        *
;                           |                                           *
;                           +-Unused - set to 0 			*
;                                                                       *
; Description:                                                          *
;                                                                       *
; This routine will get current information regarding the emulator.     *
;                                                                       *
;************************************************************************

Alignment

m6502GetInfo	proc
		xor     eax, eax        ; Zero our return code

		mov     al, MINOR_REV   ; Minor revision
		mov     ah, MAJOR_REV   ; Major revision

ifdef MULTI_CPU
		or      eax, 10000h     ; Multiple CPU support
endif

ifdef TICKS
		or      eax, 20000h     ; Timer information available
endif

ifdef SYNC_WANTED
		or      eax, 40000h     ; Sync emulation enabled
endif

ifdef BOUNDS_CHECK
		or      eax, 80000h     ; Bounds checking enabled
endif

if CYCLE_METHOD EQ INST_COUNT
		or	eax, 100000h	; We're counting instructions
endif
		ret

m6502GetInfo	endp

;************************************************************************
;                                                                       *
; Name  : m6502exec_(instructions)                                      *
;                                                                       *
; Entry : # Of 6502 instructions to execute.                            *
;                                                                       *
; Exit  : 010000h If Execution was valid                                *
;         010001h If attempted access outside of 64K area.              *
;        <010000h If invalid instruction - address of invalid instr.    *
;                                                                       *
; Description:                                                          *
;                                                                       *
; This routine will cause the virtual 6502 code to be emulated. Read the*
; header notes at the beginning of this file for programming            *
; information on setting up the 6502 for execution.                     *
;                                                                       *
;************************************************************************

Alignment

m6502exec	proc	cycles:DWORD

		mov	eax, cycles

		push    ebp
		push    ebx                     ; Save all registers we use
		push    ecx
		push    edx
		push    esi
		push    edi

		mov	cyclesRemaining, eax	; Store our cycles remaining

		cld                             ; Go forward!

		xor     eax, eax                ; Zero EAX 'cause we use it!
		xor     ebx, ebx                ; Zero EBX, too
		xor     ecx, ecx                ; Zero ECX
		xor     edx, edx                ; And EDX
		xor     edi, edi                ; Zero EDI as well
		xor     esi, esi                ; Zero our source address

		mov     bl, m6502x             ; Get our X value
		mov     cl, m6502y             ; And our Y value
		mov     dx, m6502af            ; Accumulator & flags
		mov     si, m6502pc            ; Get our program counter
		mov     ebp, m6502ram         ; Get the base address for ram
		mov     edi, m6502rom         ; Get the base address for rom

      naturalizeESI
;;;		add     esi, edi                ; Add in our base address

      ExecuteNextInstruction 0

; We get to invalidInsByte if it's a single byte invalid opcode

invalidInsByte:
		dec     esi                     ; Back up one instruction...
      push esi
      denaturalizeESI
      mov eax,esi
      pop esi		; an invalid instruction
		jmp     short emulateEnd

instReturn:
		mov     eax, 010000h            ; Indicate successful exec

; Now let's tuck away the virtual registers for next time

emulateEnd:

		mov     m6502x, bl             ; Store X
		mov     m6502y, cl             ; Store Y
		mov     m6502af, dx            ; Store A & flags

      denaturalizeESI
		mov     m6502pc, si            ; store pc

		pop     edi                     ; Restore registers
		pop     esi
		pop     edx
		pop     ecx
		pop     ebx
		pop     ebp

		ret

OutOfRange:
		mov     eax, 10001h             ; Indicates a boundary
						; violation (outside 64K)
		jmp     short emulateEnd

m6502exec	endp

_TEXT           ends
		end


