//functions for implementation of memory mappers
#include <stdlib.h>
#include <string.h>


#include "nes.h"
#include "nesvideo.h"
#include "ppu.h"
#include "mmc.h"

#include "message.h"



extern int blah,blah2;
//------------------------------------------
//no memory mapper

class mmc_none:public memorymapper
{
 public:

 //read from VRAM like normal
 byte ppuread(word a) {return ppu->read(a);}

 virtual void reset()
 {
  //copy last page of rom to end of cpu address space
  memcpy(ram+0xC000,ROM[numrom-1],sizeof(ROMPAGE));

  //copy first page of rom to 0x8000
  if (numrom>=2)  memcpy(ram+0x8000,ROM,sizeof(ROMPAGE));

   //copy vrom to ppu address space (if it exists)
  if (numvrom>0) memcpy(ppu,VROM,sizeof(VROMPAGE));
 }
};


//------------------------------------------
//Konami memory mapper

class mmc_konami:public mmc_none
{
 byte rombank; //last 16k rom bank set
 int bankswitches; //count of bank switches

 public:
 mmc_konami() {} //{mirroring=VMIRROR;} //?????
 void reset()
 {
  mmc_none::reset();
  rombank=0; bankswitches=0;
  mirroring=VMIRROR;
  msg.printf(1,"Konami mapper reset");
 }

  //write to ROM area
 int write(word a,byte d)
  {
   if (d<numrom)
    {
     if (d==rombank) return 0; //it's already set
     rombank=d;
     memcpy(ram+0x8000,&ROM[rombank],sizeof(ROMPAGE));
     bankswitches++;
    } else msg.error("KONAMI[%X]=%X",a,d);
   return 0;
  }
};


//------------------------------------------
//VROMswitch memory mapper

class mmc_vromswitch:public mmc_none
{
 char vrombank; //current 8K vrom bank
 char oldvrombank;

 int bankswitches; //count of bank switches
 public:
 mmc_vromswitch() {}

 //copy vrombank to ppu address space
 void copyVROM()
 {
  memcpy(ppu,&VROM[vrombank],sizeof(VROMPAGE)); //copy vrom
  nv->resetpatterncache(); //reset pattern cache
  oldvrombank=vrombank;
  bankswitches++;
 }
 //reset mapper
 void reset()
 {
  mmc_none::reset();
  vrombank=0; bankswitches=0;
  copyVROM();
  msg.printf(2,"VROMswitch reset");
 }
 //write
 virtual int write(word a,byte d)
 {
  vrombank=d&3; //store vrombank
  return 0;
 }
 //switch vroms before data
 void doVROMswitch()
 {
  if (vrombank!=oldvrombank) copyVROM(); //copy it!
 }

 //read from VROM bank selected
 byte ppuread(word a)
  {
   if (a>=0x2000) return ppu->read(a);
   //read from VROM
   return VROM[vrombank][a];
  }
};



//------------------------------------------
//Sequential memory mapper

class mmc_sequential:public mmc_none
{
 byte reg[4]; //4 registers

 char bit;    //next bit expected
 byte tempreg; //temp reg being read

 public:
 mmc_sequential() {}

 void reset()
 {
  mmc_none::reset();
  bit=0; tempreg=0;
  reg[0]=reg[1]=reg[2]=reg[3]=0;
  msg.printf(2,"Sequential mapper reset");
 }

 void updatereg(int num,byte d)
 {
  switch(num)
  {
   case 0: //Mirroring / vrom page select
     reg[0]=d;
//     msg.printf(2,"mirror %X",d);
  {
   byte newmirroring=(reg[0]&1) ? HMIRROR : VMIRROR;
   if (newmirroring!=mirroring)
    {
     mirroring=newmirroring;
     nv->ntc[0]->totalupdate();
     nv->ntc[1]->totalupdate();
    }
  }
    break;
   case 1: //VROM page select
     reg[1]=d;
    break;
   case 2: //second VROM page select
     reg[2]=d;
    break;
   case 3: //ROM PAGE select
     memcpy(ram+0x8000,&ROM[d],sizeof(ROMPAGE));
     reg[3]=d;
    break;
  }
 }

 virtual int write(word a,byte d)
 {
//  msg.printf(1,"seq[%X]=%X",a,d);
  if (d&0x80) {bit=0; return 0; } //reset
  if (d&1) tempreg|=(1<<bit);
  bit++;
  if (bit>=5)
   {
    updatereg((a-0x8000)/0x2000,tempreg);
    bit=0; tempreg=0;
   }
  return 0;
 }
 void doVROMswitch()
  {
  }
};








//------------------------------------------
//create memory mapper
memorymapper *newmemorymapper(int type)
{
 switch (type)
 {
  case MMC_KONAMI: return new mmc_konami;
  case MMC_VROMSWITCH: return new mmc_vromswitch;
  case MMC_SEQ: return new mmc_sequential;

   //no memory mapper
  case MMC_NONE: return new mmc_none;
  default:
   msg.error("#%d: %s MMC not supported",type,getmmctypestr(type));
   return new mmc_none;
 };
}


//-------------------------------

char *banktypestr[]=
{
 "None",    //0
 "Sequential",  //1
 "Konami",   //2
 "VROM Switch", //3
 "5202 Chip", //4

 0,0,0,0,0,0,0,0,0,0,0,0,0
};

char *getmmctypestr(int type)
{
 return banktypestr[type] ? banktypestr[type] : "Unknown";
}






