//functions for implementation of memory mappers
#include <stdlib.h>
#include <string.h>


#include "nes.h"
#include "nesvideo.h"
#include "ppu.h"
#include "mmc.h"

#include "message.h"

#include "m6502.h"

#include "r2img.h"
#include "font.h"
#include "dd.h"

#include "keyb.h"

#include "timing.h"

#define FREE(x) if (x) {free(x); x=0;}
#define DELETE(x) if (x) {delete x; x=0;}

extern int blah,blah2;


//set ROM at 0x8000 to *rom
void setrombank(ROMBANK *rom)
{
 m6502rom=(byte *)(rom-1); //offset by 0x8000
}

void ROMBANK::setlowpage(ROMPAGE *p)
{
 if (!p) memset(s+0,0,sizeof(ROMPAGE)); //clear it
    else memcpy(s+0,p,sizeof(ROMPAGE)); //set it
}

void ROMBANK::sethighpage(ROMPAGE *p)
{
 if (!p) memset(s+0x4000,0,sizeof(ROMPAGE)); //clear it
    else memcpy(s+0x4000,p,sizeof(ROMPAGE)); //set it
}


inline byte NES_ppumemory::read(word a)
{
 if (a>=0x2000 && a<0x3000) //name table wants to be read
    return nv->ntc[(a-0x2000)/0x400]->read(a&0x3FF);
  else
    return ((byte *)this)[a]; //pattern table or palette....
}



//------------------------------------------
//no memory mapper

class mmc_none:public memorymapper
{
 ROMBANK r; //only one rombank

 public:

 //read from VRAM like normal
 byte ppuread(word a) {return ppu->read(a);}

 virtual void reset()
 {
  //initialize rombank
  r.setlowpage(&ROM[0]);
  r.sethighpage(&ROM[numrom-1]);

  //set it!
  setrombank(&r);

   //copy vrom to ppu address space (if it exists)
  if (numvrom>0) memcpy(ppu,VROM,sizeof(VROMPAGE));
 }
};





//------------------------------------------
//Konami memory mapper

class mmc_konami:public memorymapper
{
 ROMBANK *r; //list of all possible rom bank combinations

 byte rombank; //last 16k rom bank set

 public:
 mmc_konami()
  {VBLANKLINES=24;
   r=(ROMBANK *)malloc(numrom*sizeof(ROMBANK));} //create all rom banks
 ~mmc_konami() {free(r);}

 void save(char *t) {t[0]=rombank;}
 void restore(char *t) {rombank=t[0]; setrombank(&r[rombank]);}

 void reset()
 {
  //copy rom into rom banks
  for (int i=0; i<numrom; i++)
   {
    r[i].setlowpage(&ROM[i]);
    r[i].sethighpage(&ROM[numrom-1]);
   }
  setrombank(&r[0]);

  rombank=0;
//  msg.printf(1,"Konami mapper reset");
 }

  //write to ROM area
 int write(word a,byte d)
  {
//msg.error("KONAMI[%X]=%X",a,d);
   //switch rom banks
   if (d<numrom)
    {
  //   if (d==rombank) return 0; //it's already set
     setrombank(&r[d]);
     rombank=d;
//     memcpy(ram+0x8000,&ROM[rombank],sizeof(ROMPAGE));
     return 1;
    }
   return 0;
  }

 virtual byte getbank() {return rombank;}

 byte ppuread(word a) {return ppu->read(a);}
};


//------------------------------------------
//VROMswitch memory mapper

int getscanline();

class mmc_vromswitch:public mmc_none
{
 char vrombank; //current 8K vrom bank
 char oldvrombank;

 int bankswitches; //count of bank switches
 public:
 mmc_vromswitch() {}

 //copy vrombank to ppu address space
 void copyVROM()
 {
  memcpy(ppu,&VROM[vrombank],sizeof(VROMPAGE)); //copy vrom
  nv->ptn[0]->reset(); //reset pattern cache
  nv->ptn[1]->reset(); //reset pattern cache
  oldvrombank=vrombank;
  bankswitches++;
 }

 void save(char *t) {t[0]=vrombank;}
 void restore(char *t) {vrombank=t[0]; copyVROM();}

 //reset mapper
 void reset()
 {
  mmc_none::reset();
  vrombank=0; bankswitches=0;
  copyVROM();
//  msg.printf(2,"VROMswitch reset");
 }
 //write
 virtual int write(word a,byte d)
 {
  //if (d<numvrom)
  if (numvrom) vrombank=d%numvrom; //store vrombank
//  msg.printf(2,"vrombank[%X]=%X line=%d",a,d%numvrom,getscanline());
  return 0;
 }
 //switch vroms before data
 void doVROMswitch()
 {
  if (vrombank!=oldvrombank) copyVROM(); //copy it!
 }

 //read from VROM bank selected
 byte ppuread(word a)
  {
   if (a>=0x2000) return ppu->read(a);
   //read from VROM
   return VROM[vrombank][a];
  }
};


//------------------------------------------
//Sequential memory mapper

class mmc_sequential:public memorymapper
{
 ROMBANK *r; //list of all possible rom bank combinations
 byte reg[4]; //4 registers

 byte vromupdated; //has vrom changed?
 byte vrompage[2]; //the vrom page settings

 byte numvrom8k; //number of 8k vrom pages
 byte numvrom4k; //number of 4k vrom pages

 char bit;    //next bit expected
 byte tempreg; //temp reg being read

 public:
 mmc_sequential() {r=(ROMBANK *)malloc(numrom*sizeof(ROMBANK));}
 ~mmc_sequential() {free(r);}

 virtual byte getbank() {return reg[3];}

 void reset()
 {
  //copy rom into rom banks
  for (int i=0; i<numrom; i++)
   {
    r[i].setlowpage(&ROM[i]);
    r[i].sethighpage(&ROM[numrom-1]);
   }
  setrombank(&r[0]);

   //copy vrom to ppu address space (if it exists)
  if (numvrom>0) memcpy(ppu,VROM,sizeof(VROMPAGE));

  numvrom8k=numvrom;
  numvrom4k=numvrom*2;

  bit=0; tempreg=0;
  reg[0]=reg[1]=reg[2]=reg[3]=0;
  vrompage[0]=vrompage[1]=0xFF;
//  msg.printf(2,"Sequential mapper reset");
 }

 void updatereg(int num,byte d)
 {
//     msg.printf(1,"seq[%X]=%X",num,d);
  switch(num)
  {
   case 0: //Mirroring / vrom page select
     if ((d&0x10)!=(reg[0]&0x10))
      {
       vrompage[0]=vrompage[1]=0xFF;
       reg[0]=reg[1]=1;
      }
     reg[0]=d;
     nv->setmirroring((reg[0]&1) ? HMIRROR : VMIRROR);
//     msg.printf(1,"seq[%X]=%X",num,d);

    break;
   case 1: //VROM page select
    if (!numvrom) break;
    reg[1]=(reg[0]&0x10) ?  (d<numvrom4k ? d : 0) : (d<numvrom8k ? d : 0);
//     if (d>0xF)
//     msg.printf(1,"seq[%X]=%X line=%d",num,d,getscanline());
    break;
   case 2: //second VROM page select
    if (!numvrom) break;
    reg[2]=(reg[0]&0x10) ?  (d<numvrom4k ? d : 0) : 0; //(d<numvrom8k ? d : 0);
//     if (d>0xF) msg.printf(1,"seq[%X]=%X",num,d);
    break;
   case 3: //ROM PAGE select
//     msg.printf(1,"seq[%X]=%X",num,d);
     if (d==reg[3]) return;
     if (d<numrom) setrombank(&r[d%numrom]);
//     memcpy(ram+0x8000+0x0000,((char *)&ROM[d])+0x0000,0x8000);
     reg[3]=d;
    // blah++;
    break;
  }
 }

 void save(char *t)
  {for (int i=0; i<4; i++)  t[i]=reg[i];}
 void restore(char *t)
  {for (int i=0; i<4; i++)  updatereg(i,t[i]);}

 virtual int write(word a,byte d)
 {
//  if (!bit) msg.printf(1,"seq[%X]=%X",a,d);
  if (d&0x80) {bit=0; tempreg=0; return 0; } //reset
  if (d&1) tempreg|=(1<<bit);
  bit++;
  if (bit>=5)
   {
    updatereg((a-0x8000)/0x2000,tempreg);
    bit=0; tempreg=0;
    return 1;
   }
  return 0;
 }
 void doVROMswitch()
  {
//   font[0]->printf(50,100,"reg[0]=%X",reg[0]);
   if (!numvrom) return;
   if (reg[0]&0x10) //2 4K ROM banks
    {
     if (vrompage[0]!=reg[1]) //page has changed
       if (reg[1]<numvrom4k)
        {
         vrompage[0]=reg[1];
         memcpy(&ppu->pt[0],((char *)VROM)+vrompage[0]*0x1000,0x1000); //copy vrom
         nv->ptn[0]->reset(); //reset pattern cache
//         blah++;
        }

     if (vrompage[1]!=reg[2]) //page has changed
       if (reg[2]<numvrom4k)
        {
         vrompage[1]=reg[2];
         memcpy(&ppu->pt[1],((char *)VROM)+vrompage[1]*0x1000,0x1000); //copy vrom
         nv->ptn[1]->reset(); //reset pattern cache
//         blah++;
        }
    } else //8k rom banks
    {
     if (vrompage[0]!=reg[1]) //page has changed
       if (reg[1]<numvrom8k)
        {
         vrompage[0]=reg[1];
         memcpy(ppu,&VROM[vrompage[0]],0x2000); //copy vrom
         nv->ptn[0]->reset(); //reset pattern cache
         nv->ptn[1]->reset();
//         blah++;
        }
    }

  }
// byte ppuread(word a) {return ppu->read(a);}


 //read from VROM bank selected
 byte ppuread(word a)
  {
   if (a>=0x2000) return ppu->read(a);
   //read from VROM
   if (reg[0]&0x10) return reg[1]<numvrom8k ? VROM[reg[1]][a] : 0; //8k
    else
     if (a<0x1000) return reg[1]<numvrom4k ? ((byte *)VROM)[reg[1]*0x1000+a] : 0;//first 4k rom
             else return reg[2]<numvrom4k ? ((byte *)VROM)[reg[2]*0x1000+a-0x1000] : 0;//first 4k rom
  }

};



//------------------------------------------
//5202 memory mapper

//rom configuration
struct rompage5202
{
 byte page[4];

 int operator ==(rompage5202 &t) {return *((unsigned *)this)==*((unsigned *)&t);}
};

//possible rom
struct ROM5202
{
 rompage5202 rp;      //page numbers for data
 unsigned refcnt;     //how many times has this been used?

 byte rom[4][0x2000]; //actual data

 //create page
 void create(rompage5202 _rp)
 {
  rp=_rp;
  refcnt=0;
  for (int i=0; i<4; i++)
   {
    if (rp.page[i]>=numrom*2) rp.page[i]=0; //ignore it
    memcpy(rom[i],((char *)ROM)+rp.page[i]*0x2000,0x2000); //copy page!
   }
//  msg.printf(1,"%X %X %X %X",rp.page[0],rp.page[1],rp.page[2],rp.page[3]);
 }

  //set this rom bank as the current one
 void set()
  {
   refcnt++; setrombank((ROMBANK *)&rom);
//  msg.printf(1,"%X %X %X %X",rp.page[0],rp.page[1],rp.page[2],rp.page[3]);
  }
};

#define NUMROMCACHE 32

extern byte IRQline,doIRQ;
class mmc_5202:public memorymapper
{
 int numrom8k;  //number of 8k roms
 int numvrom1k; //number of 1k vroms

 byte vrompage[8]; //current 1k vrom pages set
 byte oldvrompage[8]; //last 1k vrom pages set

 rompage5202 rp; //rom page config
 ROM5202  *R[NUMROMCACHE]; //cache of rom pages

 byte command; //last command read
 public:
 mmc_5202() {HBLANKCYCLES=150; memset(R,0,NUMROMCACHE*sizeof(ROM5202 *));}
 ~mmc_5202() {for (int i=0; i<NUMROMCACHE; i++) DELETE(R[i]);}

 void setrompages()
 {
  int i;
  for (i=0; i<NUMROMCACHE; i++)
   if (R[i] && R[i]->rp==rp)//does one already exist?
      {R[i]->set(); return;}  //set it

  //one was not found...create one
  for (i=0; i<NUMROMCACHE; i++)
   if (!R[i]) {R[i]=new ROM5202; break;}

  //cache is full! find the least used one
  if (i==NUMROMCACHE)
   {
    unsigned minrefcnt=0xFFFFFFFF;
    byte minpage=0;
    for (i=0; i<NUMROMCACHE; i++)
     if (R[i]->refcnt<minrefcnt) {minrefcnt=R[i]->refcnt; minpage=i;}
    i=minpage; //this is the page to use
   }

  //fill page up
  R[i]->create(rp);
  R[i]->set();
 }

 void reset()
 {
  numrom8k=numrom*2;
  numvrom1k=numvrom*8;

  //set rom page stuff
  rp.page[0]=numrom8k-2;
  rp.page[1]=numrom8k-1;
  rp.page[2]=numrom8k-2;
  rp.page[3]=numrom8k-1;
  setrompages();

  //set all vrom page stuff
  for (int i=0; i<8; i++) {vrompage[i]=oldvrompage[i]=i;}
  if (numvrom>0) memcpy(ppu,VROM,sizeof(VROMPAGE));

  command=0;
  //msg.printf(2,"5202 mapper reset");
 }

  //do command c with data d
 int docommand(byte c,byte d)
 {
  //if (c<6)
//  msg.printf(2,"command[%X]=%X",c,d);
  byte vs=(c&0x80) ? 4 : 0;
//  byte rs=(c&0x40) ? 1 : 0;
  switch (c&7)
   {
    case 0: d&=~1; vrompage[0^vs]=d; vrompage[1^vs]=(d|1); return 0;
    case 1: d&=~1; vrompage[2^vs]=d; vrompage[3^vs]=(d|1); return 0;
    case 2: vrompage[4^vs]=d; return 0;
    case 3: vrompage[5^vs]=d; return 0;
    case 4: vrompage[6^vs]=d; return 0;
    case 5: vrompage[7^vs]=d; return 0;
    case 6: rp.page[(c&0x40) ? 2 : 0]=d; setrompages(); return 1;
    case 7: rp.page[(c&0x40) ? 1 : 1]=d; setrompages(); return 1;
    default:
     msg.error("unrecognized command %X",c);
     return 0;
   }
 }

 virtual int write(word a,byte d)
 {
//  if ((a&0xFF00)!=0x8000) msg.printf(2,"5202[%X]=%X pc=%X",a,d,m6502pc);
  switch (a)
  {
   case 0xA000:
    nv->setmirroring(d ? HMIRROR : VMIRROR);
   return 0;

   case 0x8000: command=d; return 0;
   case 0x8001: return docommand(command,d);
  case 0xE000:
//    msg.printf(2,"5202[%X]=%X",a,d);
   return 0;

   case 0xC000:
   case 0xC001:
   case 0xE001:
    doIRQ=1;
    IRQline=d;
//    msg.printf(3,"5202[%X]=%X",a,d);
//    if (d&0x20) //do questionable swap
//     msg.printf(1,"before: %X %X %X %X",rp.page[0],rp.page[1],rp.page[2],rp.page[3]);
/*     {
      byte temp=rp.page[1];
      rp.page[1]=rp.page[2];
      rp.page[2]=temp;
      setrompages();
//      msg.printf(1,"SWAP");
     }*/
//     msg.printf(1,"after: %X %X %X %X",rp.page[0],rp.page[1],rp.page[2],rp.page[3]);
   return 0;

   default:  return 0;
  }
 }

 void doVROMswitch()
 {
  for (int i=0; i<8; i++)
   if (vrompage[i]!=oldvrompage[i])
    {
//     msg.printf(2,"vrompage[%X]=%X",i,vrompage[i]);
     if (vrompage[i]>=numvrom1k) vrompage[i]=numvrom1k-1;
//     memcpy( ((char *)ppu)+i*0x400,((char *)VROM)+vrompage[i]*0x400,0x400); //copy 1k bank
     for (int j=0; j<0x400; j++)
       ppu->write(i*0x400+j,*(((char *)VROM)+vrompage[i]*0x400+j));
//     nv->ptn[i/4]->reset();
     oldvrompage[i]=vrompage[i];
//     blah++;
    }
 }

 void save(char *t)
  {
   int i;
   for (i=0; i<4; i++) t[i]=rp.page[i];
   for (i=0; i<8; i++) t[i+4]=vrompage[i];
   t[12]=nv->mirroring;
  }
 void restore(char *t)
  {
   int i;
   for (i=0; i<4; i++) rp.page[i]=t[i];
   for (i=0; i<8; i++) vrompage[i]=t[i+4];
   setrompages();
   nv->setmirroring(t[12]);
  }
/*
 virtual int keyhit(char kbscan, char key)
 {
  static byte vp;
  static byte num=0;
  if (kbscan>=KB_F1 && kbscan<=KB_F8) //switch vp to change
   {vp=kbscan-KB_F1; num=0; return 1;}

   if (key>='0' && key<='9')
    {
     num*=0x10;
     num|=(key-'0');
     return 1;
    }
   if (key>='A' && key<='F')
    {
     num*=0x10;
     num|=(key-'A'+10);
     return 1;
    }
   if (kbscan==KB_ENTER)
    {
     vrompage[vp]=num;
     num=0;
     return 1;
    }

  return 0;
 }*/

 byte ppuread(word a)
 {
  if (a>=0x2000) return ppu->read(a);
  byte vp=vrompage[(a/0x400)];
  return ((char *)VROM)[(vp*0x400)+(a&0x3FF)];
 }

};




//------------------------------------------
//create memory mapper
memorymapper *newmemorymapper(int type)
{
 switch (type)
 {
  case MMC_KONAMI: return new mmc_konami;
  case MMC_SEQ: return new mmc_sequential;

  case MMC_VROMSWITCH: return new mmc_vromswitch;
  case MMC_NONE: return new mmc_none;
  case MMC_5202: return new mmc_5202;

  default:
   msg.error("#%d: %s MMC not supported",type,getmmctypestr(type));
   return new mmc_none;
 };
}


//-------------------------------

char *banktypestr[]=
{
 "None",    //0
 "Sequential",  //1
 "Konami",   //2
 "VROM Switch", //3
 "5202 Chip", //4
 "Castlevania 3", //5

 0,0,0,0,0,0,0,0,0,0,0,0,0
};

char *getmmctypestr(int type)
{
 return banktypestr[type] ? banktypestr[type] : "Unknown";
}





/*
class mmc_none:public memorymapper
{

 public:

 //read from VRAM like normal
 byte ppuread(word a) {return ppu->read(a);}

 virtual void reset()
 {
  //copy last page of rom to end of cpu address space
  memcpy(ram+0xC000,ROM[numrom-1],sizeof(ROMPAGE));

  //copy first page of rom to 0x8000
  if (numrom>=2)  memcpy(ram+0x8000,ROM,sizeof(ROMPAGE));

   //copy vrom to ppu address space (if it exists)
  if (numvrom>0) memcpy(ppu,VROM,sizeof(VROMPAGE));

  //set it!
  setrombank((ROMBANK *)(ram+0x8000));

  msg.printf(1,"Linear-MMC-none mapper reset");
 }
};*/

/*
class mmc_sequential:public memorymapper
{
 ROMBANK *r; //list of all possible rom bank combinations
 byte reg[4]; //4 registers

 byte vromupdated; //has vrom changed?
 byte vrompage[2]; //the vrom page settings

 byte numvrom8k; //number of 8k vrom pages
 byte numvrom4k; //number of 4k vrom pages

 char bit;    //next bit expected
 byte tempreg; //temp reg being read

 public:
 mmc_sequential() {r=(ROMBANK *)malloc(numrom*sizeof(ROMBANK));}
 ~mmc_sequential() {free(r);}

 virtual byte getbank() {return reg[3];}

 void reset()
 {

  //copy last page of rom to end of cpu address space
  memcpy(ram+0xC000,ROM[numrom-1],sizeof(ROMPAGE));

  //copy first page of rom to 0x8000
  if (numrom>=2)  memcpy(ram+0x8000,ROM,sizeof(ROMPAGE));

   //copy vrom to ppu address space (if it exists)
  if (numvrom>0) memcpy(ppu,VROM,sizeof(VROMPAGE));

  //set it!
  setrombank((ROMBANK *)(ram+0x8000));

  numvrom8k=numvrom;
  numvrom4k=numvrom*2;

  bit=0; tempreg=0;
  reg[0]=reg[1]=reg[2]=reg[3]=0;
  vrompage[0]=vrompage[1]=0xFF;
  msg.printf(2,"Fake-Sequential mapper reset");
 }

 void updatereg(int num,byte d)
 {
//     msg.printf(1,"seq[%X]=%X",num,d);
  switch(num)
  {
   case 0: //Mirroring / vrom page select
     if ((d&0x10)!=(reg[0]&0x10))
      {
       vrompage[0]=vrompage[1]=0xFF;
       reg[0]=reg[1]=1;
      }
     reg[0]=d;
     nv->setmirroring((reg[0]&1) ? HMIRROR : VMIRROR);
    break;
   case 1: //VROM page select
    if (!numvrom) break;
    reg[1]=(reg[0]&0x10) ?  (d<numvrom4k ? d : 0) : (d<numvrom8k ? d : 0);
//     if (d>0xF) msg.printf(1,"seq[%X]=%X",num,d);
    break;
   case 2: //second VROM page select
    if (!numvrom) break;
    reg[2]=(reg[0]&0x10) ?  (d<numvrom4k ? d : 0) : 0; //(d<numvrom8k ? d : 0);
//     if (d>0xF) msg.printf(1,"seq[%X]=%X",num,d);
    break;
   case 3: //ROM PAGE select
//     msg.printf(1,"seq[%X]=%X",num,d);
     //if (d<numrom) setrombank(&r[d%numrom]);
     if (d==reg[3]) return;
     memcpy(ram+0x8000,&ROM[d],sizeof(ROMPAGE));

     reg[3]=d;
    // blah++;
    break;
  }
 }

 virtual int write(word a,byte d)
 {
//  if (!bit) msg.printf(1,"seq[%X]=%X",a,d);
  if (d&0x80) {bit=0; tempreg=0; return 0; } //reset
  if (d&1) tempreg|=(1<<bit);
  bit++;
  if (bit>=5)
   {
    updatereg((a-0x8000)/0x2000,tempreg);
    bit=0; tempreg=0;
    return 1;
   }
  return 0;
 }
 void doVROMswitch()
  {
//   font[0]->printf(50,100,"reg[0]=%X",reg[0]);
   if (!numvrom) return;
   if (reg[0]&0x10) //2 4K ROM banks
    {
     if (vrompage[0]!=reg[1]) //page has changed
       if (reg[1]<numvrom4k)
        {
         vrompage[0]=reg[1];
         memcpy(&ppu->pt[0],((char *)VROM)+vrompage[0]*0x1000,0x1000); //copy vrom
         nv->resetpatterncache(); //reset pattern cache
         blah++;
        }

     if (vrompage[1]!=reg[2]) //page has changed
       if (reg[2]<numvrom4k)
        {
         vrompage[1]=reg[2];
         memcpy(&ppu->pt[1],((char *)VROM)+vrompage[1]*0x1000,0x1000); //copy vrom
         nv->resetpatterncache(); //reset pattern cache
         blah++;
        }
    } else //8k rom banks
    {
     if (vrompage[0]!=reg[1]) //page has changed
       if (reg[1]<numvrom8k)
        {
         vrompage[0]=reg[1];
         memcpy(ppu,&VROM[vrompage[0]],0x2000); //copy vrom
         nv->resetpatterncache(); //reset pattern cache
         blah++;
        }
    }

  }
  int keyhit(char kbscan,char key)
  {
   if (key>='0' && key<='9')
    {
     int num=key-'0';
     memcpy(&ppu->pt[0],((char *)VROM)+num*0x800,0x800); //copy vrom
     nv->resetpatterncache(); //reset pattern cache
     return 1;
    }
   if (key>='A' && key<='F')
    {
     int num=key-'A'+10;
     memcpy(&ppu->pt[0],((char *)VROM)+num*0x800,0x800); //copy vrom
     nv->resetpatterncache(); //reset pattern cache
    // reg[1]=key-'A'+10; doVROMswitch();
     return 1;
    }
   return 0;
  }

 //read from VROM bank selected
 byte ppuread(word a)
  {
   if (a>=0x2000) return ppu->read(a);
   //read from VROM
   if (reg[0]&0x10) return reg[1]<numvrom8k ? VROM[reg[1]][a] : 0; //8k
    else
     if (a<0x1000) return reg[1]<numvrom4k ? ((byte *)VROM)[reg[1]*0x1000+a] : 0;//first 4k rom
             else return reg[2]<numvrom4k ? ((byte *)VROM)[reg[2]*0x1000+a-0x1000] : 0;//first 4k rom
  }
};

  */


/*
class mmc_konami:public memorymapper
{
 char rombank;
 public:

 void reset()
 {
  //copy last page of rom to end of cpu address space
  memcpy(ram+0xC000,ROM[numrom-1],sizeof(ROMPAGE));

  //copy first page of rom to 0x8000
  if (numrom>=2)  memcpy(ram+0x8000,ROM,sizeof(ROMPAGE));

   //copy vrom to ppu address space (if it exists)
  if (numvrom>0) memcpy(ppu,VROM,sizeof(VROMPAGE));

  //set it!
  setrombank((ROMBANK *)(ram+0x8000));

  msg.printf(1,"Fake-Konami mapper reset");
  rombank=0;
 }

  //write to ROM area
 int write(word a,byte d)
  {
   if (d<numrom)
    {
     if (d==rombank) return 0; //it's already set
     rombank=d;
     memcpy(ram+0x8000,&ROM[rombank],sizeof(ROMPAGE));
    } else  msg.error("KONAMI[%X]=%X",a,d);

   return 0;
  }

 byte ppuread(word a) {return ppu->read(a);}
};*/

