//functions for implementation of memory mappers
#include <stdlib.h>
#include <string.h>


#include "nes.h"
#include "nesvideo.h"
#include "ppu.h"
#include "mmc.h"

#include "message.h"

#include "m6502.h"

#include "r2img.h"
#include "font.h"
#include "dd.h"

extern int blah,blah2;


//set ROM at 0x8000 to *rom
void setrombank(ROMBANK *rom)
{
 m6502rom=(byte *)(rom-1); //offset by 0x8000
}

void ROMBANK::setlowpage(ROMPAGE *p)
{
 if (!p) memset(s+0,0,sizeof(ROMPAGE)); //clear it
    else memcpy(s+0,p,sizeof(ROMPAGE)); //set it
}

void ROMBANK::sethighpage(ROMPAGE *p)
{
 if (!p) memset(s+0x4000,0,sizeof(ROMPAGE)); //clear it
    else memcpy(s+0x4000,p,sizeof(ROMPAGE)); //set it
}


inline byte NES_ppumemory::read(word a)
{
 if (a>=0x2000 && a<0x3000) //name table wants to be read
    return nv->ntc[(a-0x2000)/0x400]->read(a&0x3FF);
  else
    return ((byte *)this)[a]; //pattern table or palette....
}



//------------------------------------------
//no memory mapper

class mmc_none:public memorymapper
{
 ROMBANK r; //only one rombank

 public:

 //read from VRAM like normal
 byte ppuread(word a) {return ppu->read(a);}

 virtual void reset()
 {
  //initialize rombank
  r.setlowpage(&ROM[0]);
  r.sethighpage(&ROM[numrom-1]);

  //set it!
  setrombank(&r);

   //copy vrom to ppu address space (if it exists)
  if (numvrom>0) memcpy(ppu,VROM,sizeof(VROMPAGE));
 }
};



/*
class mmc_none:public memorymapper
{

 public:

 //read from VRAM like normal
 byte ppuread(word a) {return ppu->read(a);}

 virtual void reset()
 {
  //copy last page of rom to end of cpu address space
  memcpy(ram+0xC000,ROM[numrom-1],sizeof(ROMPAGE));

  //copy first page of rom to 0x8000
  if (numrom>=2)  memcpy(ram+0x8000,ROM,sizeof(ROMPAGE));

   //copy vrom to ppu address space (if it exists)
  if (numvrom>0) memcpy(ppu,VROM,sizeof(VROMPAGE));

  //set it!
  setrombank((ROMBANK *)(ram+0x8000));

  msg.printf(1,"Linear-MMC-none mapper reset");
 }
};*/


//------------------------------------------
//Konami memory mapper

class mmc_konami:public memorymapper
{
 ROMBANK *r; //list of all possible rom bank combinations

 byte rombank; //last 16k rom bank set
 int bankswitches; //count of bank switches

 public:
 mmc_konami() {r=(ROMBANK *)malloc(numrom*sizeof(ROMBANK));} //create all rom banks
 ~mmc_konami() {free(r);}

 void reset()
 {
  //copy rom into rom banks
  for (int i=0; i<numrom; i++)
   {
    r[i].setlowpage(&ROM[i]);
    r[i].sethighpage(&ROM[numrom-1]);
   }
  setrombank(&r[0]);

  rombank=0; bankswitches=0;
  msg.printf(1,"Konami mapper reset");
 }

  //write to ROM area
 int write(word a,byte d)
  {
// msg.error("KONAMI[%X]=%X",a,d);
   //switch rom banks
   if (d<numrom)
    {
     if (d==rombank) return 0; //it's already set
     setrombank(&r[d]);
     rombank=d;
     bankswitches++;
     return 1;
    }
   return 0;
  }

 byte ppuread(word a) {return ppu->read(a);}
};



/*
class mmc_konami:public memorymapper
{
 char rombank;
 public:

 void reset()
 {
  //copy last page of rom to end of cpu address space
  memcpy(ram+0xC000,ROM[numrom-1],sizeof(ROMPAGE));

  //copy first page of rom to 0x8000
  if (numrom>=2)  memcpy(ram+0x8000,ROM,sizeof(ROMPAGE));

   //copy vrom to ppu address space (if it exists)
  if (numvrom>0) memcpy(ppu,VROM,sizeof(VROMPAGE));

  //set it!
  setrombank((ROMBANK *)(ram+0x8000));

  msg.printf(1,"Fake-Konami mapper reset");
  rombank=0;
 }

  //write to ROM area
 int write(word a,byte d)
  {
   if (d<numrom)
    {
     if (d==rombank) return 0; //it's already set
     rombank=d;
     memcpy(ram+0x8000,&ROM[rombank],sizeof(ROMPAGE));
    } else  msg.error("KONAMI[%X]=%X",a,d);

   return 0;
  }

 byte ppuread(word a) {return ppu->read(a);}
};*/



//------------------------------------------
//VROMswitch memory mapper

class mmc_vromswitch:public mmc_none
{
 char vrombank; //current 8K vrom bank
 char oldvrombank;

 int bankswitches; //count of bank switches
 public:
 mmc_vromswitch() {}

 //copy vrombank to ppu address space
 void copyVROM()
 {
  memcpy(ppu,&VROM[vrombank],sizeof(VROMPAGE)); //copy vrom
  nv->resetpatterncache(); //reset pattern cache
  oldvrombank=vrombank;
  bankswitches++;
  blah++;
 }
 //reset mapper
 void reset()
 {
  mmc_none::reset();
  vrombank=0; bankswitches=0;
  copyVROM();
  msg.printf(2,"VROMswitch reset");
 }
 //write
 virtual int write(word a,byte d)
 {
  vrombank=d&3; //store vrombank
  nv->setmirroring((d&2) ? HMIRROR : VMIRROR);
  //msg.printf(2,"vrombank[%X]=%X",a,d);
  return 0;
 }
 //switch vroms before data
 void doVROMswitch()
 {
  if (vrombank!=oldvrombank) copyVROM(); //copy it!
 }

 //read from VROM bank selected
 byte ppuread(word a)
  {
   if (a>=0x2000) return ppu->read(a);
   //read from VROM
   return VROM[vrombank][a];
  }
};





//------------------------------------------
//Sequential memory mapper

class mmc_sequential:public memorymapper
{
 ROMBANK *r; //list of all possible rom bank combinations
 byte reg[4]; //4 registers

 byte vromupdated; //has vrom changed?
 byte vrompage[2]; //the vrom page settings

 byte numvrom8k; //number of 8k vrom pages
 byte numvrom4k; //number of 4k vrom pages

 char bit;    //next bit expected
 byte tempreg; //temp reg being read

 public:
 mmc_sequential() {r=(ROMBANK *)malloc(numrom*sizeof(ROMBANK));}
 ~mmc_sequential() {free(r);}

 void reset()
 {
  //copy rom into rom banks
  for (int i=0; i<numrom; i++)
   {
    r[i].setlowpage(&ROM[i]);
    r[i].sethighpage(&ROM[numrom-1]);
   }
  setrombank(&r[0]);

   //copy vrom to ppu address space (if it exists)
  if (numvrom>0) memcpy(ppu,VROM,sizeof(VROMPAGE));

  numvrom8k=numvrom;
  numvrom4k=numvrom*2;

  bit=0; tempreg=0;
  reg[0]=reg[1]=reg[2]=reg[3]=0;
  vrompage[0]=vrompage[1]=0xFF;
  msg.printf(2,"Sequential mapper reset");
 }

 void updatereg(int num,byte d)
 {
//     msg.printf(1,"seq[%X]=%X",num,d);
  switch(num)
  {
   case 0: //Mirroring / vrom page select
     if ((d&0x10)!=(reg[0]&0x10))
      {
       vrompage[0]=vrompage[1]=0xFF;
       reg[0]=reg[1]=1;
      }
     reg[0]=d;
     nv->setmirroring((reg[0]&1) ? HMIRROR : VMIRROR);
    break;
   case 1: //VROM page select
    if (!numvrom) break;
    reg[1]=(reg[0]&0x10) ?  (d<numvrom4k ? d : 0) : (d<numvrom8k ? d : 0);
//     if (d>0xF) msg.printf(1,"seq[%X]=%X",num,d);
    break;
   case 2: //second VROM page select
    if (!numvrom) break;
    reg[2]=(reg[0]&0x10) ?  (d<numvrom4k ? d : 0) : 0; //(d<numvrom8k ? d : 0);
//     if (d>0xF) msg.printf(1,"seq[%X]=%X",num,d);
    break;
   case 3: //ROM PAGE select
//     msg.printf(1,"seq[%X]=%X",num,d);
     if (d<numrom) setrombank(&r[d]);
     reg[3]=d;
    // blah++;
    break;
  }
 }

 virtual int write(word a,byte d)
 {
//  if (!bit) msg.printf(1,"seq[%X]=%X",a,d);
  if (d&0x80) {bit=0; tempreg=0; return 0; } //reset
  if (d&1) tempreg|=(1<<bit);
  bit++;
  if (bit>=5)
   {
    updatereg((a-0x8000)/0x2000,tempreg);
    bit=0; tempreg=0;
    return 1;
   }
  return 0;
 }
 void doVROMswitch()
  {
//   font[0]->printf(50,100,"reg[0]=%X",reg[0]);
   if (!numvrom) return;
   if (reg[0]&0x10) //2 4K ROM banks
    {
     if (vrompage[0]!=reg[1]) //page has changed
       if (reg[1]<numvrom4k)
        {
         vrompage[0]=reg[1];
         memcpy(&ppu->pt[0],((char *)VROM)+vrompage[0]*0x1000,0x1000); //copy vrom
         nv->resetpatterncache(); //reset pattern cache
         blah++;
        }

     if (vrompage[1]!=reg[2]) //page has changed
       if (reg[2]<numvrom4k)
        {
         vrompage[1]=reg[2];
         memcpy(&ppu->pt[1],((char *)VROM)+vrompage[1]*0x1000,0x1000); //copy vrom
         nv->resetpatterncache(); //reset pattern cache
         blah++;
        }
    } else //8k rom banks
    {
     if (vrompage[0]!=reg[1]) //page has changed
       if (reg[1]<numvrom8k)
        {
         vrompage[0]=reg[1];
         memcpy(ppu,&VROM[vrompage[0]],0x2000); //copy vrom
         nv->resetpatterncache(); //reset pattern cache
         blah++;
        }
    }

  }
// byte ppuread(word a) {return ppu->read(a);}


 //read from VROM bank selected
 byte ppuread(word a)
  {
   if (a>=0x2000) return ppu->read(a);
   //read from VROM
   if (reg[0]&0x10) return reg[1]<numvrom8k ? VROM[reg[1]][a] : 0; //8k
    else
     if (a<0x1000) return reg[1]<numvrom4k ? ((byte *)VROM)[reg[1]*0x1000+a] : 0;//first 4k rom
             else return reg[2]<numvrom4k ? ((byte *)VROM)[reg[2]*0x1000+a-0x1000] : 0;//first 4k rom
  }

  int keyhit(char kbscan,char key)
  {
/*   if (key>='0' && key<='9')
    {
     int num=key-'0';
     memcpy(&ppu->pt[0],((char *)VROM)+num*0x800,0x800); //copy vrom
     nv->resetpatterncache(); //reset pattern cache
     return 1;
    }
   if (key>='A' && key<='F')
    {
     int num=key-'A'+10;
     memcpy(&ppu->pt[0],((char *)VROM)+num*0x800,0x800); //copy vrom
     nv->resetpatterncache(); //reset pattern cache
    // reg[1]=key-'A'+10; doVROMswitch();
     return 1;
    }*/
   return 0;
  }
};




//------------------------------------------
//create memory mapper
memorymapper *newmemorymapper(int type)
{
 switch (type)
 {
  case MMC_KONAMI: return new mmc_konami;
  case MMC_SEQ: return new mmc_sequential;

  case MMC_VROMSWITCH: return new mmc_vromswitch;
  case MMC_NONE: return new mmc_none;

  default:
   msg.error("#%d: %s MMC not supported",type,getmmctypestr(type));
   return new mmc_none;
 };
}


//-------------------------------

char *banktypestr[]=
{
 "None",    //0
 "Sequential",  //1
 "Konami",   //2
 "VROM Switch", //3
 "5202 Chip", //4

 0,0,0,0,0,0,0,0,0,0,0,0,0
};

char *getmmctypestr(int type)
{
 return banktypestr[type] ? banktypestr[type] : "Unknown";
}






