//emulation of NES registers....
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <process.h>

#include "types.h"

#include "message.h"
#include "file.h"

#include "keyb.h"

#define FREE(x) if (x) {free(x); x=0;}
#define DELETE(x) if (x) {delete x; x=0;}

#include "m6502\m6502.h"
#include "timing.h"

#include "nesvideo.h"
#include "nes.h"
#include "mmc.h"

#include "input.h"
#include "dd.h"

#include "r2img.h"
#include "font.h"

//extern struct m6502MemoryWrite NESwrite[];
//extern struct m6502MemoryRead NESread[];
extern int blah,blah2;

WORD NESread(word a);
int NESwrite(word a,byte d);


//---------------------------------
//hardware internal registers
word ppumemaddr; //address for next write in ppu memory
char ppureadfuckup;  //first mem fuck up

byte spritememaddr; //next write in sprite memory

volatile byte vblank; //are we in a vblank ?
volatile byte hitflag; //are we in hitflag?
volatile byte frame; //are we scanning frame?

byte mirroring; //ppu name table mirroring

memorymapper *mmc; //current memory mapper

//-------------------------
//reset nes

//create data structures for nes hardware
void initNEShardware()
{
 if (NESinitialized) return;
 ppu=new NES_ppumemory;

 ram=(char *)malloc(0x10000);
 msg.printf(2,"64K CPU address space created");

 spritemem=(NES_sprite *)malloc(256);
 msg.printf(2,"256 byte sprite mem created");

 mmc=newmemorymapper(romhdr.banktype); //create memorymapper
 NESinitialized=1;
}

//free data structures for nes hardware
void terminateNEShardware()
{
 if (!NESinitialized) return;
 m_stop();
 DELETE(ppu);
 DELETE(mmc);
 FREE(ram);
 FREE(spritemem);
 NESinitialized=0;
}

int resetNEShardware()
{
 if (!NESinitialized) return -1;
 sl.clear();
 ppumemaddr=0; ppureadfuckup=0;
 spritememaddr=0;

 //erase memory
 ppu->clear();
 memset(ram,0,0x10000);
 memset(spritemem,0,256);

 //set default mirroring
 mirroring=romhdr.mirroring;

 if (!numrom) {msg.error("No rom data!"); return -1;}

 //reset memory mapper
 mmc->reset();

 //set up 6502 CPU emulator
 m6502Base=(uchar *)ram; //set base of emu
 m6502reset();   //reset cpu

 //set r/w trap handlers
 m6502MemoryRead=NESread;
 m6502MemoryWrite=NESwrite;

 //reset emulated video
 nv->resetpalette();
 nv->resetpatterncache();


 CPUpaused=0;
 msg.printf(1,"NES hardware reset");
 return 0;
}


//--------------------------------
//PPU emulation

NES_ppumemory::NES_ppumemory()
{
 clear();
 msg.printf(2,"%dK PPU address space created",sizeof(*this)/1024);
}

void NES_ppumemory::clear()
{
 memset(this,0,sizeof(*this));
}

void NES_ppumemory::write(word a,byte d)
{
//msg.printf(1,"ppu data write %X %X",ppumemaddr,d);

 //pallete write ?
 if (a>=0x3F00 && a<0x3F20)
  {
//   msg.printf(1,"pal[%X]=%X",a-0x3f00,d);
   if (((byte *)this)[a]==d) return; //no change in palette
   if (!(a&0xF)) //it's a background palette...
    {
     for (int i=0; i<8; i++)
      {
       ((byte *)this)[0x3F00+i*4]=d;
        nv->palupdateidx[i*4]=1;
      }
     nv->paletteupdated=1;
    } else       //normal palette
    if (a&3)
    {
     ((byte *)this)[a]=d; //write it then...
     nv->palupdateidx[a-0x3F00]=1;
     nv->paletteupdated=1;
    }
   return;
  }


 //pattern table was written to ?
 if (a<0x2000)
 {
  if (((byte *)this)[a]==d) return; //no change in
 ((byte *)this)[a]=d; //write byte to pattern table
  nv->ptnupdateidx[a/0x1000][(a&0xFFF)/sizeof(NES_pattern)]=1;
  nv->patternupdated=1;
//  blah2++;
//  msg.printf(1,"pattern %d.%d updated",a/0x1000,(a&0xFFF)/sizeof(NES_pattern));
  return;
 }

 //name/attribute table was written to..
 if (a<0x3000)
  {
//msg.printf(1,"nametable write[%X]=%X frame=%d",a,d,frame);
//  if (((byte *)this)[a]==d) return; //no change in
    //we must mirror it
   if (romhdr.mirroring==HMIRROR)
    {
     ((byte *)this)[a]=d; //write byte to ppu memory
     ((byte *)this)[a^0x400]=d; //horizontal mirroring
      nv->ntc[(a&0x800) ? 1 : 0]->update(a&0x3FF);
      //nv->ntc[0]->update(a&0x3FF);
     // nv->ntc[1]->update(a&0x3FF);
    }
   else
    {
     ((byte *)this)[a]=d; //write byte to ppu memory
     ((byte *)this)[a^0x800]=d; //vertical mirroring
     nv->ntc[(a&0x400) ? 1 : 0]->update(a&0x3FF);
    }
    return;
  }

}




//---------------------------------------


void scrolllist::reset()
{
 flip=0; //reset flip
 defaultsp.natable=lastnatable;
 defaultsp.line=0;

 spnum=0; //empty list
 sp[spnum]=defaultsp; //store default
 naline=-1;
// msg.printf(2,"reset");
}

void scrolllist::writenatable(byte d)
 {
// msg.printf(2,"na=%d",d);
  int line=getscanline();
//  if (line>naline)
   {
//    msg.printf(2,"naline=%d line=%d na=%d frame=%d KEEP ",naline,line,d,frame);
    sp[spnum].natable=d;
    lastnatable=d;
    naline=line;
   } //else    msg.printf(2,"naline=%d line=%d na=%d frame=%d IGNORE ",naline,line,d,frame);
 }

void scrolllist::add() {sp[spnum++]=s;}
void scrolllist::writescroll(byte d)
{
 //msg.printf(2,"line=%d val=%d",getscanline(),d);
 if (!flip)  s.x=d;   //read x
  else
  {
   if (d<=239) s.y=d; //read y
      else msg.error("yover");
   s.line=getscanline(); //store scanline
//   s.natable=lastnatable; //sp[spnum].natable; //currentnatable;
   add();     //add it to end array
//   s.print();
   s.reset();
  }
 flip^=1;
}

//find heights of all blocks
void scrolllist::resolve(int yw)
{
 //create default scroll pos for next time..
 if (!spnum) spnum++;

 //resolve all heights
 sp[spnum].line=yw;
 int i;
 for (i=0; i<spnum; i++)
  if (sp[i].line) sp[i].line-=2;
 for (i=0; i<spnum; i++)
 {
  sp[i].height=sp[i+1].line-sp[i].line; //find height of bank
//  if (sp[i].height && i<spnum-1) sp[i].height++;
 }
 //set last scrollpos to last written natable
 sp[spnum-1].natable=lastnatable;
 defaultsp=sp[spnum-1];
}

void scrollpos::print()
{
 msg.printf(1,"pos=%d,%d line=%d natable=%d",x,y,line,natable);
}


//-----------------------------------------

void startframe()
{
 frame=1;     //during scanning of frame
 vblank=0;    //not in vblank
}

void startvblank()
{
 vblank=1;   //we are in vblank
 hitflag=0;  //reset hitflag
 frame=0;
}


//-------------------------------------------------------
//memory write trap handlers

//write to lower ports
int neswrite2000(WORD a,BYTE d)
{
  switch (a)
  {
   case 0x2000: //PPU control register 1
//   msg.printf(1,"write[%X]=%X",a,d);
//     if ((d&4)!=(ram[a]&4))
//      msg.printf(1,"%d writes",(d&4) ? 32 : 1);

      sl.writenatable(d&3);
      ram[a]=d;
      return 0;
   case 0x2001: //PPU control register 2
      ram[a]=d;
      return 0;
   case 0x2003: //sprite memory address
      spritememaddr=d;
//      msg.printf(1,"sprite mem addr set to %X",spritememaddr);
      return 0;
   case 0x2004: //sprite mem data write
//      msg.printf(1,"spritewrite %d",spritememaddr);
      ((byte *)spritemem)[spritememaddr++]=d;
      return 0;
   case 0x2005: //Background scroll
      sl.writescroll(d); //write to scroll list
     return 0;

   case 0x2006: //PPU memory address
     ram[0x2006]^=1;
     ((byte *)&ppumemaddr)[ram[0x2006]]=d;
     ppureadfuckup=1;
//     if (!ram[0x2006]) msg.printf(1,"ppu mem addr set to %X frame=%d",ppumemaddr,frame);
      return 0;
   case 0x2007: //PPU data write
      ppu->write(ppumemaddr,d);
      ppumemaddr+=(ram[0x2000]&4) ? 32 : 1;
      return 0;
   default:
    msg.printf(1,"unsupported write: %X %X",a,d);
    return 0;
  }
};

//write to upper ports
int neswrite4000(WORD a,BYTE d)
{
// msg.error("upper port write!!");
 switch (a)
 {
  case 0x4014: //DMA sprite copy
//     msg.printf(3,"DMA sprite copy from %X",d*0x100);
     memcpy(spritemem,ram+(d*0x100),0x100); //copy memory
    return 0;
  case 0x4015: //sound switch
     ram[0x4015]=d;
    return 0;
  case 0x4016: //joystick
  case 0x4017: //joystick
     ram[0x4016]=1; //reset joystick
     ram[0x4017]=1; //reset joystick
    return 0;
  default:
//    msg.printf(1,"unsupported write: %X %X",a,d);
   return 0;
 };
}

//write to expansion module
int neswrite5000(WORD a,BYTE d)
{
 msg.error("expansion write!!");
 return 0;
}




//-------------------------------------------------------
//memory read trap handlers

WORD nesread2000(WORD a)
{
//  msg.printf(3,"%X read",a);
  switch (a)
  {
   case 0x2000: //PPU control register 1
    msg.printf(3,"%X read",a);
      return ram[a];
   case 0x2001: //PPU control register 2
    msg.printf(3,"%X read",a);
      return ram[a];
   case 0x2002: //PPU status register
    {
      byte s=0;
      if (vblank) {s|=0x80; vblank=0;} //reset vblank after read

        //see if we've scanned past sprite #0's y coordinate
      if (!hitflag)
      {
//       msg.printf(2,"hit! line=%d hitline=%d",getscanline(),spritemem[0].y-1);
       if (spritemem[0].y <= getscanline())
        {s|=0x40; hitflag=1;}
      }
//      msg.printf(2,"PPU status read %X",s);
     return s;
    }

   case 0x2004: //sprite mem data read
//      msg.printf(1,"spriteread %d",spritememaddr);
      return ((byte *)spritemem)[spritememaddr++];
   case 0x2005: msg.error("bgscroll read!");  return 0;
   case 0x2007: //PPU data read
     {
//      if (ppureadfuckup) {ppureadfuckup=0; return ppu->read(ppumemaddr);}
//      byte d=ppu->read(ppumemaddr);
//      ppumemaddr+=(ram[0x2000]&4) ? 32 : 1;

      byte d=mmc->ppuread(ppumemaddr); //use memory mapper
      if (ppureadfuckup) {ppureadfuckup=0; return d;}
      ppumemaddr+=(ram[0x2000]&4) ? 32 : 1;
      return d;
     }
   default: return 0;
  }
}

WORD nesread4000(WORD a)
{
// msg.error("upper port read!!");
 switch (a)
 {
  case 0x4016: //joystick
  case 0x4017:
   {
    int joynum=a-0x4016; //get joystick number
    int stat=inputdevice[joynum]->stat; //get current status of dirs
    int but=inputdevice[joynum]->but; //get current status of buts
    switch (ram[a]++)
    {
     case 1: return (but&ID_BUT1) ? 1 : 0; //A
     case 2: return (but&ID_BUT0) ? 1 : 0; //B
     case 3: return (but&ID_BUT2) ? 1 : 0; //Select
     case 4: return (but&ID_BUT3) ? 1 : 0; //Start
     case 5: return (stat&ID_UP ) ? 1 : 0; //
     case 6: return (stat&ID_DOWN) ? 1 : 0; //
     case 7: return (stat&ID_LEFT) ? 1 : 0; //
     case 8: return (stat&ID_RIGHT) ? 1 : 0; //
    }
   }
 }
 return 0;
}

WORD nesread5000(WORD a)
{
 msg.error("expansion read!!");
 return 0;
}


//------------------------------------
//main read and write handlers

WORD NESread(word a)
{
 if (a<0x4000) return nesread2000(a); //lower ports
 if (a<0x5000) return nesread4000(a); //higher ports
 return 0; //unsupported
}


int NESwrite(word a,byte d)
{
 if (a>=0x8000) return mmc->write(a,d); //redirect to memory mapper
 if (a<0x4000) return neswrite2000(a,d); //lower ports
 if (a<0x5000) return neswrite4000(a,d); //higher ports
 return 0; //unsupported
}














