//functions for NES display
#include <string.h>
#include <stdlib.h>

#include "r2img.h"
#include "font.h"

#include "nesvideo.h"
#include "nes.h"
#include "mmc.h"
#include "slist.h"

#include "gui.h"
#include "message.h"

#include "config.h"

#include "dd.h"

#include "keyb.h"

#define FREE(x) if (x) {free(x); x=0;}
#define DELETE(x) if (x) {delete x; x=0;}


//color index from which all nes colors are based
#define CBASE 224

nesvideo *nv; //global nes video

#include "file.h"


#define min(a,b) (((a)<(b)) ? (a) : (b))
#define max(a,b) (((a)>(b)) ? (a) : (b))

extern int blah,blah2;


//--------------------------
//palette stuff

#include "nespal.h"

void initializenespalette()
{
/* FILEIO f;
 f.open("d:\\emu\\ines\\nesa.com");
 f.setpos(0xB4e);
 for (int i=0; i<0x40; i++)
  {
   char c;
   f.read(&nespal[i],3); //read palette
   f.read(&c,1);
   msg.printf(2,"%X: %X %X %X",i,nespal[i].r,nespal[i].g,nespal[i].b);
  }
 f.close();

 f.create("nes.pal");
 f.write(nespal,256*3);
 f.close();*/

 FILEIO f;
 f.open("anes.pal");
 f.read(nespal,256*3);
 f.close();

// nespal[0x27].r=0xfc;
// nespal[0x27].g=0x7c;
// nespal[0x27].b=0x5c; //3c
/*
 nespal[0x37].r=0xfc;
 nespal[0x37].g=0xbc;
 nespal[0x37].b=0x7c;*/

 msg.printf(2,"NES Palette initialized");
}

//32 different indexes
void updatepalette(int palidx,byte paldata)
{
 setpalettenum(CBASE+palidx,&nespal[paldata]);
//  msg.printf(1,"palette[%d]=%X (%X,%X,%X)",palidx,paldata,
//     nespal[paldata].r,nespal[paldata].g,nespal[paldata].b);
}

void nesvideo::refreshpalette()
{
 if (!paletteupdated) return;
 //update bg pals
 for (int i=0; i<32; i++)
  if (palupdateidx[i])
   {
    updatepalette(i,ppu->bgpal.c[i]);
    palupdateidx[i]=0;
   }
 paletteupdated=0;
}

void nesvideo::resetpalette()
{
 for (int i=0; i<32; i++) palupdateidx[i]=1;
 paletteupdated=1;
}

//--------------------------------
//pattern stuff

byte bitmap8x8::create(NES_pattern *np)
{
 memset(s,0,8*8); //clear it
 byte issolid=0;
 //create 256 color pattern out of nes bit pattern
 for (int y=0; y<8; y++)
  for (int x=0; x<8; x++)
   {
    if (np->low[y] &(0x80>>x)) s[y][x]|=1;
    if (np->high[y]&(0x80>>x)) s[y][x]|=2;
    issolid|=np->low[y]|np->high[y];
   }
 return !issolid;
}

void bitmap8x8::add_sprite(byte b)
{
 char *t=(char *)s;
 for (int i=0; i<64; i++,t++)
  if (*t) *t|=b;
}

void bitmap8x8::add_tile(byte b)
{
 char *t=(char *)s;
 for (int i=0; i<64; i++,t++) *t|=b;
}



//create pattern from NES_pattern
void pattern::refresh()
{
 bitmap8x8 b;
 issolid=b.create(np); //create bitmap

// if (issolid) msg.printf(2,"solid tile!");
 //create bg tiles
 for (int i=0; i<4; i++)
  {
   tile[i]=b; //copy bitmap
   tile[i].add_tile(CBASE+(i<<2)); //add attribute and cbase
  }

 //create sprites
 for (i=0; i<4; i++)
  {
   sprite[i]=b; //copy bitmap
   sprite[i].add_sprite(CBASE+16+(i<<2)); //add attribute and cbase
  }
}

inline void pattern::setpatternptr(int tablenum,int index)
{
 np=&ppu->pt[tablenum].p[index];
}

//---------------------------------
//pattern table cache

void patterntable::reset()
{ //reset patterns
 memset(ptnupdateidx,1,256);
 updated=1;
}

void patterntable::init(int ptnum)
{
 //set pattern pointers
 for (int i=0; i<256; i++)
  ptn[i].setpatternptr(ptnum,i);
 reset();
}

//update only tiles that have changed in patterntable
void natablecache::individualupdate(byte *pu)
{
 for (int y=0; y<30; y++)
  for (int x=0; x<32; x++)
   if (pu[nt.nt.t[y][x]]) updated[y][x]=1;
 doanyupdate=1;   
}

void patterntable::refresh()
{
 if (!updated) return;

 //update patterns
 for (int j=0; j<256; j++)
  if (ptnupdateidx[j]) ptn[j].refresh();

 //see if nametable needs updating as well
 if (this==nv->realntc[0]->p)  nv->realntc[0]->individualupdate(ptnupdateidx);
 if (this==nv->realntc[1]->p)  nv->realntc[1]->individualupdate(ptnupdateidx);

 //clear updates
 memset(ptnupdateidx,0,256);
 updated=0;
}




//------------------------------
//NES_natable
seventlist sl;
scontextlist sc;

//rotation lookup
byte attriblookup[4][4];

void initattriblookup()
{
 for (int x=0; x<4; x++)
  for (int y=0; y<4; y++)
   {
    byte square=(x>>1) | (y&2);
    attriblookup[x][y]=square*2;
   }
}

inline byte NES_attributetable::getat(int x,int y)
{
 return (a[y/4][x/4]>>attriblookup[x&3][y&3])&3;
}

inline void NES_natable::drawtile(char *dest,int tx,int ty,struct pattern *p)
{
  p[nt.t[ty][tx]].tile[at.getat(tx,ty)].draw_tile(dest,tx*8,ty*8);
   blah++;

}

//draws a name table at a x,y (will do the clipping)
void NES_natable::draw(char *dest,int sx,int sy,pattern *p)
{
 int xstart,xend;
 int ystart,yend;

 if (sx<0) {xstart=(-sx)/8; xend=32;}//left clip
    else   {xstart=0; xend=min((SCREENX-sx)/8+1,32); }
 if (sy<0) {ystart=(-sy)/8; yend=30;}//top clip
    else   {ystart=0; yend=min((SCREENY-sy)/8+1,30);}

 for (int y=ystart; y<yend; y++)
  for (int x=xstart; x<xend; x++)
   p[nt.t[y][x]].tile[at.getat(x,y)].draw_tile(dest,sx+x*8,sy+y*8);
}


//----------------------------------------------
//natable cache

void natablecache::createsurface() {s=new surface(32*8,30*8); totalupdate();}
void natablecache::freesurface() {delete s;}
natablecache::natablecache():p(0) {createsurface();}
natablecache::~natablecache() {freesurface();}

void natablecache::refresh()
{
 if (!doanyupdate || !p) return; //no updating needed

 //lock surface
 char *dest=s->lock();
 if (!dest) return;

 if (dototalupdate)
  {
   memset(updated,0,32*30); //clear update table
   nt.draw(dest,0,0,p->ptn);
   dototalupdate=0;
  } else
  {
   for (int y=0; y<30; y++)
    for (int x=0; x<32; x++)
     if (updated[y][x]) //this tile # has been changed
     {
      nt.drawtile(dest,x,y,p->ptn); //redraw individual tile to cache
      updated[y][x]=0;
     }
  }

 //unlock surface
 s->unlock();
 doanyupdate=0;
}


//draw to screen at x,y clipped to clip
void natablecache::draw(int x,int y,lrect *clip)
{
 if (!s->blt(
  max(0,clip->x1-x), //x coord within source
  max(0,clip->y1-y), //y coord within source

  min(32*8,clip->x2-x), //x2 coord
  min(30*8,clip->y2-y), //y2
  screen,
  max(x,clip->x1),max(y,clip->y1)  //destination
  ))  totalupdate();
}


//------------------------------------------
//nesvideo draw bg

//(x,y)-(x2,y2) is the region to draw bg to
void nesvideo::drawbg(char *dest,int x,int y,int x2,int y2,byte sprites)
{
  sl.print(50,50);
 sc.print(200,50);
// return;
 //create clipping rectangle
 lrect clip;
 //set left/right clipping
 clip.x1=x; clip.x2=x2;

 y-=8;
 //go through all sections
 for (int i=0; i<sc.scnum; i++)
  if (sc.sc[i].height>0 && sprites==sc.sc[i].sprites) //if section has a height
  {
   clip.y1=y+max(sc.sc[i].line,8);
   clip.y2=min(y+sc.sc[i].line+sc.sc[i].height,y2);

   //scroll positions
   int sx=x-sc.sc[i].sx;
   int sy=y-sc.sc[i].sy; //for top clipping

   int natable=sc.sc[i].natable&3;
   ntc[natable]  ->draw(sx,sy,&clip);
   ntc[natable^1]->draw(sx+32*8,sy,&clip);
   ntc[natable^2]->draw(sx,sy+30*8,&clip);
   ntc[natable^3]->draw(sx+32*8,sy+30*8,&clip);
 }

//font[0]->printf(50,50,"scroll=%d,%d",scrollx,scrolly);
//font[0]->printf(50,60,"nametable=%d %s",ram[0x2000]&3,
//  nv->mirroring==VMIRROR ? "vmirror" : "hmirror");
//font[0]->printf(50,60,"ram[0x2000]=%X",ram[0x2000]);
//font[0]->printf(50,50,"ram[0x2001]=%X",ram[0x2001]);
//font[0]->printf(50,50,"%d",sl.num);
// sl.print(50,50);
// sc.print(200,50);
// int px=50,py=50;
}


//------------------------------------------------------------
// Sprites

extern int blah;
inline void NES_sprite::draw_8x8(char *dest,pattern *pt)
{
 if (x<SCREENX && y<SCREENY-8)// && p!=255)//&& !behindbg)
  if (!pt[p].issolid)
   pt[p].sprite[attrib].draw_sprite(dest,x,y+1-8,flipx|(flipy<<1),behindbg);
}

inline void NES_sprite::draw_8x16(char *dest)
{
 if (x<SCREENX && y<SCREENY-8)// && p!=255) // && !behindbg)
 {
  int pnum=p&(~1)^flipy;
  pattern *pt=nv->ptn[p&1]->ptn; 

  if (!pt[pnum].issolid)
      pt[pnum].  sprite[attrib].draw_sprite(dest,x,y+1-8,flipx|(flipy<<1),behindbg);
  pnum^=1;
  if (!pt[pnum].issolid)
       pt[pnum].sprite[attrib].draw_sprite(dest,x,y+9-8,flipx|(flipy<<1),behindbg);
 }
}


void nesvideo::drawsprites_8x8(char *dest)
{
 //find sprite pattern table to show...
 pattern *p=nv->ptn[(ram[0x2000]&8) ? 1 : 0]->ptn;

 //draw every sprite
// for (int i=0; i<64; i++)
 for (int i=63; i>=0; i--)
  spritemem[i].draw_8x8(dest,p);

font[0]->printf(50,200,"y=%d",spritemem[0].y-1);
//font[0]->printf(50,100,"flipx=%d flipy=%d",spritemem[0].flipx,spritemem[0].flipy);
}

void nesvideo::drawsprites_8x16(char *dest)
{
 //draw every sprite
// for (int i=0; i<64; i++)
 for (int i=63; i>=0; i--)
  spritemem[i].draw_8x16(dest);

  font[0]->printf(50,200,"y=%d",spritemem[0].y-1);
//font[0]->printf(50,100,"unknown=%d",spritemem[0].unknown);
//font[0]->printf(50,100,"flipx=%d flipy=%d",spritemem[0].flipx,spritemem[0].flipy);
//font[0]->printf(50,60,"ptable=%d",(ram[0x2000]&8) ? 1 : 0);
//font[0]->printf(50,50,"%d,%d",scrollx,scrolly);
}


void m_spritedump(){}


//---------------------------------
//main video interface

void cleardesktop()
{
 #ifdef WIN95
 if (nv) nv->forcedesktopfill=16;
 #else
 if (nv) nv->forcedesktopfill=1;
 #endif
}

void nesvideo::resetmirroring()
{
 if (mirroring==VMIRROR)
  {
   ntc[0]=realntc[0];
   ntc[1]=realntc[1];
   ntc[2]=realntc[0];
   ntc[3]=realntc[1];
//   msg.printf(1,"Vertical mirroring set");
  } else
  {
   ntc[0]=realntc[0];
   ntc[1]=realntc[0];
   ntc[2]=realntc[1];
   ntc[3]=realntc[1];
//   msg.printf(1,"Horizontal mirroring set");
  }
}

void nesvideo::setmirroring(int type)
{
 if (mirroring==type) return; //ignore it
 mirroring=type;
 resetmirroring();
}


void nesvideo::createnatablecaches()
{
 //create name/attribute table caches
 realntc[0]=new natablecache();
 realntc[1]=new natablecache();
 msg.printf(2,"Name table caches created: %dK",32*8*30*8/1024);
 resetmirroring();
}

void nesvideo::freenatablecaches()
{
 DELETE(realntc[0]);
 DELETE(realntc[1]);
 msg.printf(2,"Name table caches freed");
}

void nesvideo::createpatterntables()
{
 ptn[0]=new patterntable(0);
 ptn[1]=new patterntable(1);
 msg.printf(2,"Pattern table caches created");
}

void nesvideo::freepatterntables()
{
 delete ptn[0];
 delete ptn[1];
 msg.printf(2,"Pattern table caches freed");
}


void nesvideo::reset()
{
 resetpalette();
 ptn[0]->reset();
 ptn[1]->reset();
 realntc[0]->totalupdate();
 realntc[1]->totalupdate();
}


static DLGPOS dlgpos;
nesvideo::nesvideo(char *romfile):
 GUIcontents(256,224),maximized(0)
{
 if (nv) delete nv->parent; //delete old video

 xw=width(); yw=height();

 setmirroring(romhdr.mirroring);

 initializenespalette();
 initattriblookup();
 createpatterntables();
 createnatablecaches();

 //wrap us up in box
 GUImaximizebox *p=new GUImaximizebox(guiroot,romfile,this,0,0);
 dlgpos.open(p);
 if (SCREENX<=320)   p->maximize();

 msg.printf(2,"NES video initialized: %dx%d",xw,yw);
 nv=this;
 cleardesktop();
}

nesvideo::~nesvideo()
{
 cfg->set(CFG_NOFILLEDDESKTOP,0); //force desktop to be drawn

 freenatablecaches();
 freepatterntables();
 msg.printf(2,"NES video destroyed");
 nv=0;

 freerom();
 dlgpos.close((GUIbox *)parent);
}

int ccx,ccy;
int docolorcheck;

extern volatile byte inemu;

void viewcolor(byte colornum);

extern int guienabled;
//update screen
void nesvideo::draw(char *dest)
{
 cfg->set(CFG_NOFILLEDDESKTOP,maximized);

 //see if vrom needs switching before we start drawring.....
 if (CPURunning || CPUpaused) mmc->doVROMswitch();


 //update patterntable caches
 ptn[0]->refresh();
 ptn[1]->refresh();

 //update name caches' pattern tables
 patterntable *pt=nv->ptn[(ram[0x2000]&16) ? 1 : 0];
 realntc[0]->setpatterntable(pt);
 realntc[1]->setpatterntable(pt);

 //update nametable caches
 realntc[0]->refresh();
 realntc[1]->refresh();

 if (forcedesktopfill>0) {fill(0); forcedesktopfill--;}
 else //blank screen...
 if (!(ram[0x2001]&8)) {fill(0); return; }

 //find x,y coords of where to draw nes screen
 lrect r;
 r.x1=x1+(width()-xw)/2; if (r.x1<0) r.x1=0;
 r.y1=y1+(height()-yw)/2; if (r.y1<0) r.y1=0;
 r.x2=r.x1+xw;
 r.y2=r.y1+yw;

 //erase edges
 if (guienabled)
  {
   drawrect(dest,0,x1,y1,width(),r.y1-y1);   //top
   drawrect(dest,0,x1,r.y2,width(),y2-r.y2); //bottom
   drawrect(dest,0,x1,r.y1,r.x1-x1,r.y2-r.y1);   //left
   drawrect(dest,0,r.x2,r.y1,x2-r.x2,r.y2-r.y1); //right
  }

 //draw background that should contain sprites
 drawbg(dest,r.x1,r.y1,r.x2,r.y2,1);

 {
  //force clipping
  CLIP clip(dest,r.x1,r.y1,r.x2,r.y2);

  //draw sprites
//  if ((ram[0x2001]&16)) //show sprites
   if (ram[0x2000]&32) //8x16 sprites ?
         drawsprites_8x16(dest);
   else  drawsprites_8x8(dest);
 }

 //draw background that should NOT contain sprites
 drawbg(dest,r.x1,r.y1,r.x2,r.y2,0);


 if (docolorcheck)
 {
  viewcolor(dest[ccy*PITCH+ccx]-CBASE);
  docolorcheck=0;
 }
}

int nesvideo::keyhit(char kbscan,char key)
{
 return mmc ? mmc->keyhit(kbscan,key) : 0;
}

int isviewcoloropen();

#include "mouse.h"
GUIrect * nesvideo::click(mouse &m)
{
 ccx=m.x-x1;
 ccy=m.y-y1;

 //color check
 if (kbstat&KB_SHIFT)
  {
   int x=ccx,y=ccy+8;
   int ysize=(ram[0x2000]&32) ? 16 : 8;
   for (int i=0; i<64; i++)
   {
    NES_sprite *s=&spritemem[i];
    if (x>=s->x && x<s->x+8 && y>=s->y && y<s->y+ysize)
      { //sprite hit!
        msg.printf(2,"Sprite #%d: (%d,%d) pattern: %X attrib: %d",i,s->x,s->y,s->p,s->attrib);
       return 0;
      }
    }
   return 0;
  }

 if ((kbstat&KB_CTRL) || isviewcoloropen()) {docolorcheck=1; return 0;}

 return 0;
}





//--------------------------
//GUI related bullshit
void enablegui();
void nesvideo::restore()
 {
//msg.printf(2,"nesvideo restore");
  cfg->set(CFG_NOFILLEDDESKTOP,0); //force desktop to be drawn
  maximized=0;
  resize(xw,yw);
  enablegui();
  cleardesktop();
 }

void nesvideo::maximize()
 {
//msg.printf(2,"nesvideo maximize");
  cfg->set(CFG_NOFILLEDDESKTOP,1); //force desktop to not be drawn
  moveto(0,0);
  maximized=1;
  resize(SCREENX,SCREENY);
  cleardesktop();
 }
void nesvideo::resize(int txw,int tyw)
{
//msg.printf(2,"nesvideo resize %d,%d",txw,tyw);
 GUIrect::resize(txw,tyw);
 if (maximized && parent)
   ((GUImaximizebox *)parent)->reposmaxbutton();
}
