#ifndef _NESVIDEO_
#define _NESVIDEO_

#include "guirect.h"
#include "ppu.h"

extern "C" {
 void __cdecl draw_tile_asm(struct bitmap8x8 *b,char *dest,int x,int y);
 void __cdecl draw_sprite_asm(struct bitmap8x8 *b,char *dest,int x,int y,int o);
 void __cdecl draw_sprite_behind_asm(struct bitmap8x8 *b,char *dest,int x,int y,int o);
};
#ifdef __WATCOMC__
#pragma aux draw_tile_asm modify[eax ecx edx];
#pragma aux draw_sprite_asm modify[eax ecx edx];
#endif

//8x8 bitmap (for pattern tables
struct bitmap8x8
{
 byte s[8][8];
 void draw_tile(char *dest,int x,int y)
  {draw_tile_asm(this,dest,x,y);} //draw as tile
 void draw_sprite(char *dest,int x,int y,int o,int behind) //draw as sprite (transparency and orientation)
  {
   if (!behind)  draw_sprite_asm(this,dest,x,y,o);
          else   draw_sprite_behind_asm(this,dest,x,y,o);
  } //draw as sprite

 void add_tile(byte b); //add b to every pixel in bitmap
 void add_sprite(byte b); //add b to every pixel in bitmap (except 0)

 //create 256 bitmap from nes pattern
 byte create(NES_pattern *np); //returns 1 if all transparent pixels
};

//optimized pattern
struct pattern
{
 //optimized pattern data 8x8 for tiles
 bitmap8x8 tile[4];    //one for each possible attribute

 //optimized pattern data 8x8 for tiles
 bitmap8x8 sprite[4];  //one for each possible attribute

 struct NES_pattern *np; //pointer to nes pattern that this represents
 char issolid;    //is it a solid tile of pixel 0?
 char padding[3];

 void setpatternptr(int tablenum,int index);
 void refresh(); //create optimized pattern from nes data
};


//class that creates a 32x30 surface
//and receives name table writes
//and updates surface accordingly
class natablecache
{
 class surface *s; //32x30 tile surface

 char updated[36][32]; //updated flags
 char doanyupdate;
 char dototalupdate;

 friend class nesvideo;
 friend class namedlg;

 public:
 natablecache();
 ~natablecache();

 void refresh(int nanum); //update surface (if needed)
 void totalupdate() {doanyupdate=dototalupdate=1;}
 void update(unsigned a) //update tile at address a
  {
   if (a<32*30) //it's a tile...
    updated[a/32][a&31]=1; //set update flag for this tile
   else //it's an attribute
    {
     a-=32*30; //get address into att table
     int x=(a&7)*4;
     int y=(a/8)*4;

     *((unsigned *)&updated[y][x])=0x1010101;
     *((unsigned *)&updated[y+1][x])=0x1010101;
     *((unsigned *)&updated[y+2][x])=0x1010101;
     *((unsigned *)&updated[y+3][x])=0x1010101;
    }
   doanyupdate=1;
  }
};


struct scrollpos
{
 byte x,y;  //scroll position
 byte natable; //name table to use

 unsigned line; //scroll line that this was received
 unsigned height; //height of scroll block (must be resolved at end of frame)
 void reset() {x=y=0; line=0; natable=0; height=0;}
 void print();
};

class scrolllist
{
 friend class nesvideo;
 int spnum; //number received
 scrollpos sp[16]; //list of all scroll pos's

 int naline;
 byte lastnatable;

 scrollpos s; //current scroll pos being read
 scrollpos defaultsp; //default sp
 char flip; //hardware register flip

 public:
 void reset(); //reset (beginning of each frame)
 void resolve(int yw); //resolve all heights

 void add(); //adds current scrollpos to array
 void writescroll(byte d); //data written to scroll register
 void writenatable(byte d); //data written to natable register
 void print();

 scrolllist() {spnum=0; flip=0; defaultsp.reset();};
};
extern scrolllist sl;

//class for output to GUI
class nesvideo:public GUIcontents
{
 int xw,yw; //xw and yw of video
 public:
 int maximized;

 byte paletteupdated; //has any of the pal been updated?
 byte palupdateidx[32]; //which ones were updated?
 void refreshpalette();
 void resetpalette();

 pattern ptn[2][256]; //optimized patterns
 byte ptnupdateidx[2][256]; //which ones have been updated?
 byte patternupdated;  //have any?
 void refreshpatterns();
 void resetpatterncache();
 void initpatterncache();

 natablecache *ntc[2];
 void createnatablecaches();
 void freenatablecaches();

 void drawbg(char *dest,int x,int y);
 void drawsprites_8x8(char *dest);
 void drawsprites_8x16(char *dest);

 void reset();

 char forcedesktopfill;

 void check(int x,int y);

 nesvideo(char *romfile);
 virtual ~nesvideo();

 //------------------------------------
 //gui related bullshit
 virtual int acceptfocus() {return 1;}
 virtual void losefocus() { return; } //workspace never loses focus
 virtual int keyhit(char kbscan,char key);
 virtual char *getname() {return "nesvideo";};
 virtual void restore();
 virtual void maximize();
 void resize(int xw,int yw);
 virtual void draw(char *dest);
 virtual GUIrect *nesvideo::click(mouse &m);
};

void updatepalette(int palidx,byte paldata); //update palette number #

//current nes video
extern nesvideo *nv;

#endif