#ifndef _NESVIDEO_
#define _NESVIDEO_

#include "guirect.h"
#include "ppu.h"

extern "C" {
 void __cdecl draw_tile_asm(struct bitmap8x8 *b,char *dest,int x,int y);
 void __cdecl draw_sprite_asm(struct bitmap8x8 *b,char *dest,int x,int y,int o);
 void __cdecl draw_sprite_behind_asm(struct bitmap8x8 *b,char *dest,int x,int y,int o);
};
#ifdef __WATCOMC__
#pragma aux draw_tile_asm modify[eax ecx edx];
#pragma aux draw_sprite_asm modify[eax ecx edx];
#endif

//8x8 bitmap (for pattern tables
struct bitmap8x8
{
 byte s[8][8];
 inline void draw_tile(char *dest,int x,int y)
  {draw_tile_asm(this,dest,x,y);} //draw as tile
 inline void draw_sprite(char *dest,int x,int y,int o,int behind) //draw as sprite (transparency and orientation)
  {
   if (!behind)  draw_sprite_asm(this,dest,x,y,o);
          else   draw_sprite_behind_asm(this,dest,x,y,o);
  } //draw as sprite

 void add_tile(byte b); //add b to every pixel in bitmap
 void add_sprite(byte b); //add b to every pixel in bitmap (except 0)

 //create 256 bitmap from nes pattern
 byte create(NES_pattern *np); //returns 1 if all transparent pixels
};

//optimized pattern
struct pattern
{
 //optimized pattern data 8x8 for tiles
 bitmap8x8 tile[4];    //one for each possible attribute

 //optimized pattern data 8x8 for tiles
 bitmap8x8 sprite[4];  //one for each possible attribute

 struct NES_pattern *np; //pointer to nes pattern that this represents
 char issolid;    //is it a solid tile of pixel 0?
 char padding[3];

 void setpatternptr(int tablenum,int index);
 void refresh(); //create optimized pattern from nes data
};


//class that creates a 32x30 surface
//and receives name table writes
//and updates surface accordingly
class natablecache
{
 class surface *s; //32x30 tile surface

 NES_natable nt; //REAL nt
 class patterntable *p; //pattern table to use

 char updated[36][32]; //updated flags
 char doanyupdate;
 char dototalupdate;

 friend class nesvideo;
 friend class namedlg;
 friend class patterntable;
 friend class nesstate;

 public:
 natablecache();
 ~natablecache();

 void createsurface();
 void freesurface();

 void refresh(); //update surface (if needed)
 void totalupdate() {doanyupdate=dototalupdate=1;}
 void individualupdate(byte *pu); //updated based on individual pattern updates

 void setpatterntable(patterntable *newp)
  {if (p!=newp) {p=newp; totalupdate();}}

 void write(unsigned a,char d) //update tile at address a
  {
   if (((char *)&nt)[a]==d) return; //ignore it

   ((char *)&nt)[a]=d; //write byte

   //set update flags
   if (a<32*30) updated[a/32][a&31]=1; //set update flag for this tile
   else //it's an attribute
    {
     a-=32*30; //get address into att table
     int x=(a&7)*4;
     int y=(a/8)*4;

     *((unsigned *)&updated[y][x])=0x1010101;
     *((unsigned *)&updated[y+1][x])=0x1010101;
     *((unsigned *)&updated[y+2][x])=0x1010101;
     *((unsigned *)&updated[y+3][x])=0x1010101;
    }
   doanyupdate=1;
  }
  char read(unsigned a) {return ((char *)&nt)[a];}

  void draw(int x,int y, struct lrect *clip);
};

//optimized pattern table cache
class patterntable
{
 byte ptnupdateidx[256]; //which have been updated
 byte updated;  //have any?

 friend class patterntableview;
 friend class natablecache;

 public:
 pattern ptn[256];       //optimized patterns for table

 void init(int ptnum); //init cache with all patterns pointing to table ptnum
 void reset(); //reset
 void refresh();
 inline void update(unsigned a)
  {ptnupdateidx[a]=1; updated=1;}
 patterntable(int ptnum) {init(ptnum);}
};

//class for output to GUI
class nesvideo:public GUIcontents
{
 public:
 int xw,yw; //xw and yw of video
 int maximized;

 void reset();

 byte paletteupdated; //has any of the pal been updated?
 byte palupdateidx[32]; //which ones were updated?
 void refreshpalette();
 void resetpalette();

 patterntable *ptn[2]; //2 independent pattern tables
 void createpatterntables();
 void freepatterntables();

 natablecache *realntc[2]; //only 2 REAL caches
 natablecache *ntc[4];    //4 mirrored
 void createnatablecaches();
 void freenatablecaches();
 void setmirroring(int type); //set mirroring
 void resetmirroring();
 byte mirroring; //current mirroring

 void drawbg(char *dest,int x,int y,int x2,int y2,byte sprites);
 void drawsprites_8x8(char *dest);
 void drawsprites_8x16(char *dest);

 char forcedesktopfill;

 void check(char *dest,int x,int y);

 nesvideo(char *romfile);
 virtual ~nesvideo();

 //------------------------------------
 //gui related bullshit
 virtual int acceptfocus() {return 1;}
 virtual void losefocus() { return; } //workspace never loses focus
 virtual int keyhit(char kbscan,char key);
 virtual char *getname() {return "nesvideo";};
 virtual void restore();
 virtual void maximize();
 void resize(int xw,int yw);
 virtual void draw(char *dest);
 virtual GUIrect *nesvideo::click(mouse &m);
};

void updatepalette(int palidx,byte paldata); //update palette number #

//current nes video
extern nesvideo *nv;

#endif