#ifndef _SLIST_
#define _SLIST_

int getscanline();


//structure describing an event that occurs during a virtual frame
struct sevent
{
 unsigned line; //line that it occurred
 byte type; //type of event
 byte data; //actual data
// byte t[2];
// word pc;

 //event types
 #define SE_SCROLLX 1
 #define SE_SCROLLY 2
 #define SE_NATABLE 3
 #define SE_SPRITE 4

 void print(int x,int y);

 int operator ==(sevent &t) {return type==t.type && data==t.data;}
};


//list of all events that have occured during a frame
struct seventlist
{
 int num; //number of events recorded so far
 sevent s[512]; //list of events

 void reset() {num=0;} //no events
 void add(sevent &t);

 void addnatable(byte nanum)
  {
//   if (!(ram[0x2001]&8)) return;
   sevent t;
   t.type=SE_NATABLE; t.data=nanum;
   add(t);
  }
 void addscrollx(byte sx)
  {
   sevent t;
   t.type=SE_SCROLLX; t.data=sx;
   add(t);
  }
 void addscrolly(byte sy)
  {
   if (sy>239) return; //ignore
   sevent t;
   t.type=SE_SCROLLY; t.data=sy;
   add(t);
  }
 void addsprite(byte s)
  {
   sevent t;
   t.type=SE_SPRITE; t.data=s;
   add(t);
  }

  void print(int x,int y);
};



//------------------------------------
//screen context

struct scontext
{
 byte sx,sy; //scroll position
 byte natable; //name table
 byte sprites; //0-off 1-on

 unsigned line;    //top line of context
 unsigned height;  //height of context

 void clear() {sx=sy=0; natable=0; line=0; height=0;}
 void reset() {line=0;}

 void setline(int l) {line=l;} //set upper limit of context
 void setheight(int bottom) //set lower limit of context
  {height=bottom-line;}

 void operator +=(sevent &e); //add event to context
 int operator ==(scontext &t) //comparison
  {return sx==t.sx && sy==t.sy && natable==t.natable;}

 void print(int x,int y); //print
};


//list of all contexts, formed by combining all events
struct scontextlist
{
 int scnum;
 scontext c;

 scontext sc[64]; //screen contexts

 void reset() {scnum=0;}
 void clear() {scnum=0; c.clear();}
 void add(scontext &t);

 void create(seventlist &el); //create from event list

 void print(int x,int y); //print to screen
};



extern seventlist sl; //global event list
extern scontextlist sc; //global screen context


#endif








